/*<std-header incl-file-exclusion=PARAERROR_H>

           Copyright 1998, NCR Corporation.  All Rights Reserved.

                               NCR Confidential.


 $Id: paraError.h,v 1.7 1998/09/04 23:16:35 bolo Exp $


*/

#ifndef PARAERROR_H
#define PARAERROR_H


/*  -- do not edit anything above this line --   </std-header>*/

#include <string.h>
#include <iostream.h>
#include <w_strstream.h>

#include <ErrorManager.h>
#include <paraAssert.h>

#define ELOC ELoc(__FILE__, __LINE__)

// TRY(x, y)
// Call this macro to try to execute x, and on failure, include information y in the error
// report.  The original error message is not emitted, only the call location and the
// additional info.

#define PARA_TRY(x, y) \
      do { int _err = (x); \
              if (_err) { Error _e(_err); _e << "called " << ELOC << y; return _err; } \
      } while (0)

//
// ELoc
//
// This class is used to record an error location
// composed of a file name and a line number.  It provides a standard
// way of including error location information in an error message.
//
// See the ELOC macro.
//
class ELoc {
	int line;
	char *file;
public:
	ELoc(const char *f, int l) : file(0) {

		if (f) {
			file = new char[strlen(f)+1];
			strcpy(file, f);
		}
		line = l;
	};
	ELoc(const ELoc& other) : file(0) {
		line = other.line;
		if (other.file) {
			file = new char[strlen(other.file)+1];
			strcpy(file, other.file);
		}
	};

	~ELoc() {
		delete [] file;
	};
	friend inline ostream &operator<<(ostream&os, const ELoc &e);
};

//
// Shift operator for sending an error location to an Error object.
//
inline ostream &operator<<(ostream&os, const ELoc &e) {
	os << "at " << (e.file ? e.file : "<unknown>")
		 << ":" << e.line << "\t";
	return os;
};

//
// BaseError
//
// This class is the parent of all error object classes.
// The basic functionality provided here is 
//  1) an error code
//  2) a pointer to an error manager which will receive the error report
//     when the error object is destroyed.
//
// Note that the code here is the translated error code provided as a result
// of taking an error enumeration and passing it through an error map.  Error
// maps are defined in .dat files and are turned into code by Error.pl.
//
class BaseError : public ostrstream {
protected:
	int code;

	char SqlClass[3];
	char SqlSubclass[4];

	BaseErrorManager *em;
public:

	BaseError(int c = 0, BaseErrorManager *e = 0): code(c) {
		if (e == 0)
			em = BaseErrorManager::MyErrorManager();
		else
			em = e;
		SetSqlState("00", "000");
	};
	virtual ~BaseError() {
	};

	int Code() {
		return code;
	};
	const char *GetSqlClass() {
		return SqlClass;
	};
	const char *GetSqlSubclass() {
		return SqlSubclass;
	};

	operator int() {
		return Code();
	};
	void SetCode(int x) {
		code = x;
	};

	void SetSqlState(const char *cl, const char *subcl) {
		strncpy(SqlClass, cl, 2);
		strncpy(SqlSubclass, subcl, 3);
		SqlClass[2] = '\0';
		SqlSubclass[3] = '\0';
	};
};

class Error : public BaseError {

public:
	Error(int c = 0, BaseErrorManager *e = 0): BaseError(c, e) {
	};
	virtual ~Error();
};

//
// Template for creating mapped errors.  The Error object should
// be instantiated with the error code from an error map table.
//
template <class T>
class MappedError : public Error {
protected:
	static T emap;
	const char *message;
public:
	MappedError(int i=0, BaseErrorManager *e=0) : Error(emap[i].code, e) , message(emap[i].message) {
	};
};

template<class T>
BaseError &operator<<(BaseError &e, T t) {
	ostrstream &oe = e;
	oe << t;
	return e;
};

template<class T>
Error &operator<<(Error &e, T t) {
	ostrstream &oe = e;
	oe << t;
	return e;
};

class FatalError : public BaseError {
public:
	// TODO should have an enumerated list of fatal errors.
	FatalError(int c=0) : BaseError(c) {
		*this << "Fatal Error " << c << " occurred ";
	};

	void Raise() {
		*this << ends;
		BaseErrorManager::FatalError(this->str());
		// We never come back from this call.
	};
		
	virtual ~FatalError() {
		Raise();
	};
			
};


//
// UserError
// Use an instance of this class to report an error message
//
class UserError : public BaseError {

public:
	UserError(int c) : BaseError(c) {
	};

	virtual ~UserError() {
		if (code) {
			ostrstream::operator<<(endl);
			ostrstream::operator<<(ends);
			em->SetUserError(*this);
		}
#ifdef _WINDOWS 
		rdbuf()->freeze(0);
#else
		freeze(0);
#endif
	};
};

//
// Use an instance of this class to explain an error
// that has already occurred.
// E.g.
//
//  ExplainError() << "Error at line 1, invalid relation name."
//
class ExplainError : public BaseError {

public:
	ExplainError(BaseErrorManager *e = 0): BaseError(0, e) {
	};
	virtual ~ExplainError() {
		ostrstream::operator<<(endl);
		ostrstream::operator<<(ends);
		if (em == 0) {
			cerr << str() ;
#ifdef _WINDOWS
			rdbuf()->freeze(0);
#else
			freeze(0); 		// cause string to be deleted
#endif
		} else {
			em->AddExplanation(*this);
#ifdef _WINDOWS
			rdbuf()->freeze(0);
#else
			freeze(0); 		// cause string to be deleted
#endif
		}
	};
};

/*<std-footer incl-file-exclusion=PARAERROR_H>  -- do not edit anything below this line -- */

#endif          /*</std-footer>*/
