# <std-header style=TCL orig-src=PTK>
#
#            Copyright 1998, NCR Corporation.  All Rights Reserved.
#
#                                NCR Confidential.
#
#
#  $Id: metaCache.tcl,v 1.24 1998/05/19 03:22:27 kupsch Exp $
#
#  This file was originally part of ptk, a front end based on TCL/TK to the
#  Paradise database system developed by the University of Wisconsin-Madison.
#  The above copyright applies to changes made to this file by NCR Corporation.
#  The original copyright which may appear in the source below applies to the
#  original file as obtained by NCR Corporation and unchanged parts of this file.
#
#
#   -- do not edit anything above this line --   </std-header>

# ============================================================
# This file contains the following procedures to manipulate
# and access metacache information:
#
# 1. getAttrNames: return a list of attribute names of given
#	table.
# 	Arguments
#		- extName	name of the table;
#	Implementation:
#		It uses the global variabel session and attrinfo.
#
# 2. getAttrTypes: return a list of attribute types of a given
#	table.
#	Arguments
#		-extName	name of the table;
#
# 3. getFirstDispAttr: return a pair (attrname, attrtype) of the
# 	first displayable attribute for a given table.
#	Arguments
#		-extName	name of the table;
#
# 4. isAttrDispType: return a boolean indicating whether the given
#	attribute type is a displayable attribute or not.
#	Arguments
#		-typeId		ID of the type
#	Comments: the type of each attribute is stored in forms of
#		integer id in meta cache.  To convert it to names,
#		use ptkTypeid2name function.  It uses the information
#		stored in global PDtypes defined in misc.tcl
#
# 5. makePDBBox: converts a bounding box constant specified in traditional
#	diagonal format to the Paradise format: all four coordinates are
#	required in either clockwise or counterclockwise order.
#	Argument:
#		-x1 y1 x2 y2: upper-left and lower-right coordinates
#	Return: a list of coordinates of the corresponding Paradise box.
#

# ------------------------------------------------------------
proc getAttrNames extName {
	global attrinfo 

	set attrnames {}
	foreach attr $attrinfo($extName) {
		lappend attrnames [lindex $attr 0]
	}
	return $attrnames
}

# ------------------------------------------------------------
proc getAttrTypes extName {
	global attrinfo 

	set attrtypes {}
	foreach attr $attrinfo($extName) {
		lappend attrtypes [lindex $attr 1]
	}
	return $attrtypes
}

# ------------------------------------------------------------
# find first display attribute of a given extent name
# ------------------------------------------------------------
proc getFirstDispAttr extName {
	global attrinfo session

	set indx -1
	for {set i 0} {$i < [llength $attrinfo($extName)]} {incr i} {
		set attrType [lindex [lindex $attrinfo($extName) $i] 1]
		if [isAttrDispType $attrType] {
			set indx $i
			break
		}
	}
	if {$indx != -1} {
		set retval [list [lindex [lindex $attrinfo($extName) $indx] 0]				$attrType]
	} { set retval {NONE NONE} }
	return $retval
}

proc getDispAttrs extName {
	global attrinfo session
		
	set disp_attrs {}
	foreach i $attrinfo($extName) {
		set attrType [lindex $i 1]
		if [isAttrDispType $attrType] {
			lappend disp_attrs [list [lindex $i 0] $attrType]
		}
	}
	return $disp_attrs
}

proc getAllAttrs extName {
	global attrinfo session
		
	set attrs {}
	foreach i $attrinfo($extName) {
		lappend attrs [list [lindex $i 0] [lindex $i 1]]
	}
	return $attrs
}

# ------------------------------------------------------------
proc isAttrDispType typeId {
	global PDtypes

	# 0-integer; 1-real; 2-booleanAdt; 3-stringAdt; 4-pointAdt
	# 5-rectangleAdt; 6-closedPolygonAdt;
	# 7-polyLineAdt; 8-circleAdt; 9-11 Rasters; 30-swissCheese
	if {($typeId >= 4 && $typeId <= 11) || $typeId == 30} {
		return 1
	} { return 0}
}

proc isAttrHasMbrType typeId {
	global PDtypes

	# 0-integer; 1-real; 2-booleanAdt; 3-stringAdt; 4-pointAdt
	# 5-rectangleAdt; 6-closedPolygonAdt;
	# 7-polyLineAdt; 8-circleAdt; 9-11 Rasters; 30-swissCheese
	if {($typeId >= 5 && $typeId <= 11) || $typeId == 30} {
		return 1
	} { return 0}
}

#----------------------------------------------------------------------
# test whether the type id corresponds to Int4Adt or not
#----------------------------------------------------------------------
proc isAttrInt4Adt typeId {
	global PDtypes
	# 0-Int4Adt
	if {$typeId == 0} {
		return 1
	} { return 0}
}

proc isAttrRasterAdt typeId {
	global PDtypes
	if {($typeId>=9 && $typeId<=11)} {
		return 1
	} { return 0 }
}

# ------------------------------------------------------------
proc makePDBBox {x1 y1 x2 y2} {
	set newBox "ClosedPolygon (\"$x1 $y1, $x2 $y1, $x2 $y2, $x1 $y2\")"
	return $newBox
}

# ------------------------------------------------------------
proc setGridInfo sid {
	global session 

	set session($sid,xGridLine) ""
	set session($sid,xGridLineLabel) ""
	set session($sid,yGridLine) ""
	set session($sid,yGridLineLabel) ""

	set xintvr [expr $session($sid,view_w)/($session($sid,grid_x) + 1)]
	set yintvr [expr $session($sid,view_h)/($session($sid,grid_y) + 1)]
	for {set i 1; set a $xintvr} {$i <= $session($sid,grid_x)} {incr i 1} {
		lappend session($sid,xGridLine) $a; 
		lappend session($sid,xGridLineLabel) [x2laeax $a $sid]
		incr a $xintvr
	}
	for {set i 1; set a $yintvr} {$i <= $session($sid,grid_y)} {incr i 1} {
		lappend session($sid,yGridLine) $a
		lappend session($sid,yGridLineLabel) [y2laeay $a $sid]
		incr a $yintvr
	}
}

# ----------------------------------------------------------------------
proc querygen {sid extName} {
	global session debugMode

	if $debugMode { puts "querygen $sid $extName" }

	set has_raster 0

	# make a target list, the complication is that we have to
	# deal with raster attribute on target list
	set attrs [getAllAttrs $extName]
	set dispAttrs [getDispAttrs $extName]

	# check to see whether this extent can be displayed or not
	if {$dispAttrs != {}} {
		set sqlTargetList ""
		set sqlWhere ""
		foreach attr $attrs {
			set attrn [lindex $attr 0]
			set attrt [lindex $attr 1]
			if {$attrn == "oid"} {continue}

			if [isAttrDispType $attrt] {
			if [isAttrRasterAdt $attrt] {
				set has_raster 1
				append sqlTargetList \
					" $extName.$attrn.clip(##PDMAP##),"
			} else {
				append sqlTargetList " $extName.$attrn,"
				append sqlWhere \
				   " $extName.$attrn overlaps ##PDMAP## or"
			}
			} else {
				append sqlTargetList " $extName.$attrn,"
			}
		}
		if $has_raster {
			set sqlTargetList [string trimright $sqlTargetList ,]
		} else {
			# make life easier for optimizer
			set sqlTargetList " * "
		}
		set sqlWhere [string trimright $sqlWhere or]

		if {$sqlWhere == ""} {
			set SQLquery "select $sqlTargetList \
				from $extName"
		} {
			set SQLquery "select $sqlTargetList \
				from $extName where $sqlWhere"
		}
	} else {
		# no displayable attribute, set query to NONE
		set SQLquery "NONE"
	}
	if $debugMode { puts "\tgen query: $SQLquery" }
	return $SQLquery
}

#----------------------------------------------------------------------
# ptk_bbox::set
# 	set the current bounding box of the Paradise map of a
#	particular session
#	Check to see if the bbox area is 0, if so, expand it
#	so that the current point is in the center of the box
# 	This will fix the problem of zooming in too close and
#	and not able to zoom out again.  Also handle the case
# 	of centering a single point data on screen.
#----------------------------------------------------------------------
proc ptk_bbox::set {sid x1 y1 x2 y2} {
	global session
    	global BOX_CHANGE hLayerCnt
	global debugMode

	if $debugMode {
		puts "old value: x1 = $session($sid,bx1) y1 = $session($sid,by1) x2 = $session($sid,bx2) y2 = $session($sid,by2)"
		puts "new value: nx1 = $x1 ny1 = $y1 nx2 = $x2 ny2 = $y2"
	}
	set changed 0
	if {$x1 == $x2} {
	  set x1 [expr $x1 - 1];
	  set x2 [expr $x2 + 1];
	}
	if {$y1 == $y2} {
	  set y1 [expr $y1 + 1];
	  set y2 [expr $y2 - 1];
	}

	if {$session($sid,bx1) != $x1} {
		set changed 1
		set session($sid,bx1) $x1
	}
	if {$session($sid,by1) != $y1} {
		set changed 1
		set session($sid,by1) $y1
	}
	if {$session($sid,bx2) != $x2} {
		set changed 1
		set session($sid,bx2) $x2
	}
	if {$session($sid,by2) != $y2} {
		set changed 1
		set session($sid,by2) $y2
	}
	
	set x_fac [expr (double($x2 - $x1))/$session($sid,view_w)];
	set y_fac [expr (double($y2 - $y1))/$session($sid,view_h)];
	set session($sid,x_scale_fac) $x_fac
	set session($sid,y_scale_fac) $y_fac
	set session($sid,bbox_changed) $changed

	if { $changed == 1 } {
    	    foreach layer $session($sid,hidden_layers) {
		set ItemIndex [lsearch $session($sid,hidden_layers) $layer]
		puts "ItemIndex is $ItemIndex"
	    	if { $layer != {} } {
		    puts "layer is $layer"
		    set length [llength $layer]
	 	    puts "layer length is $length"
		    set layer [lreplace $layer $BOX_CHANGE $BOX_CHANGE 1]
	    	    set session($sid,hidden_layers) [lreplace \
		    $session($sid,hidden_layers) $ItemIndex $ItemIndex $layer]
		    puts "after replace, layer is $layer"
	        }
    	    }
	}
}

#
# invoked fetch all the meta info from server for the currently open
# database.  It always does a refetch because it should be called by
# a change of current open database and old cache info is no longer
# valid.
#
# global variables extnames and attrinfo and metadata will be
# loaded with corresponding values
#
proc fillMetaCache {} {
  global extnames
  global  attrinfo;
  global metadata;
  global debugMode

  # clear the contents of old array attrinfo
  if {[catch "array size attrinfo"] == 0} {
	set extnames [array names attrinfo]
	foreach i $extnames {
		unset attrinfo($i)
	}
	set attrinfo(null) {};
  }

  set metadata [PDgetextnames];
  foreach i $metadata {
    # type info in attrinfo is in terms of type id, not string name
    # if needed, convert them to string via ptkTypeid2name.
    set ext_name [lindex $i 0]
    lappend attrinfo($ext_name) [list [lindex $i 1] [lindex $i 2]];
    
  }
  if {$metadata == {}} {
	set extnames {}
  } else {
  	set extnames [array names attrinfo]
  }
  return
}

