
#include <ids_receiver.h>


//----------------------------------------------------------------------------------------
//  constructor() -  Initialize the receiver_thread.  Set up a query object and connect it
//                   to the IDS DB.  Set up the receiver socket for incoming requests
//
//  ARGS:    none
//  RETURN:  none
//----------------------------------------------------------------------------------------
idsReceiver::idsReceiver()
{
    bool con_st;
    _socket = new TCPSock(IDS_RECEIVER_PORT, con_st);
    if(con_st == false){
	cerr <<"Failed to create socket for ids receiver thread"<<endl;
	exit(1);
    }
    _query = new SQLQuery ( IDS_DB_NAME, PARADISE_PORT );
}

//----------------------------------------------------------------------------------------
//  destructor() - 
//
//  ARGS:    none
//  RETURN:  none
//----------------------------------------------------------------------------------------
idsReceiver::~idsReceiver()
{
    _socket->close();
    delete _socket;	
    end();
}


//----------------------------------------------------------------------------------------
//  run() -  Run function for this thread.  Listens on receiver_thread port and
//           responds to profile management requests.  Obtains an exclusive
//           "Writers Lock" to block other profile threads from modifying the
//           timerTable or the IDS db while profile management requests are being
//           serviced.
//
//  ARGS:    none
//
//  RETURN:  none
//----------------------------------------------------------------------------------------
int 
idsReceiver::run()
{    
    int status;
    int i;
    
    cout << "Listening on port " << IDS_RECEIVER_PORT 
	 << " for incoming client profile add/update/delete requests"<<endl;
    
    for(;;) {
	
	// Code to ensure ctrl-C and ctrl-D get caught 
	signal(SIGINT, SIG_DFL);      // catch ctrl-C 
	signal(SIGQUIT, SIG_DFL);     // catch ctrl-D      
	
	//--------------------------------------------------
	// Wait for the next profile management request
	//--------------------------------------------------
	_sendersSocket = _socket->accept();
	
INFO("IDS RECIEVER THREAD: Got Message");	
	
	//------------------------------------------------------------
        //  Open the DB, process the message, Close the DB
	//------------------------------------------------------------
	if( _query->open_db() == OK){
	    if( _sendersSocket ){
		status = processMessage();
		if( status != OK ){
		    ERROR("Error in get: error code passed back to message sender: "<<status);
		    BEM.display(status);
		}
		_sendersSocket->put( status );
		_sendersSocket->close();
	    }      
	    _query->close_db();
	}	
    }
}


//----------------------------------------------------------------------------------------
//  processMessage() - This method processes a profile management request message and responds
//                     by returning a status code to the caller
//  
//  ARGS:    none
//  RETURN:  status code { true if message processedm false if failed }
//----------------------------------------------------------------------------------------
int 
idsReceiver::processMessage()
{

    int           type_of_message;
    time_t        my_time = time(NULL);
    int           st;
    char *        recv_buf;
    char *        sql;
    char *        addr;
    int           numBytes;
    int           recvd_bytes;
    pm_message *  m;
    AProfile      ap;
    ofstream      log(IDS_RECEIVER_LOG,ios::app);

    if(!log){
	ERROR("Unable to open log file: "<<IDS_RECEIVER_LOG);
	return -1;
    }
    
    //-----------------------------------------------------------------
    //  Receive the length of the message, then get the message itself
    //-----------------------------------------------------------------
    if( _sendersSocket->get( numBytes ) == false ){
	ERROR( "Failed to get number of bytes: messageTypeCode="<< numBytes);
	return -1;
    }

    recv_buf = new char[numBytes]; 
    if(! recv_buf ){ return -1; }
    bzero(recv_buf, numBytes);

    recvd_bytes = _sendersSocket->get_bytes( recv_buf, numBytes );
    assert( recvd_bytes == numBytes );

    if(recvd_bytes != numBytes){
	ERROR( "Failed to get number of bytes: messageTypeCode="<< numBytes);
	delete recv_buf;
	return -1;
    }

    INFO("RECEIVED:  "<<recvd_bytes<<" byte message from a smartPull client");    

    m = new pm_message( recv_buf, numBytes );
    if(! m ){
	ERROR("Failed to allocate memory");
	return FAIL;
    }
    delete[] recv_buf;
    
    //------------------------------------------------------------
    //  Write the message to the log file
    //------------------------------------------------------------
    log << "//////////////////////////////////////////////////////////////////////////////////////////\n";
    log << "//  " << endl;
    log << "//  Profile Management Prequest Received at: "<< ctime(&my_time);
    log << "//  " << endl;
    log << "//////////////////////////////////////////////////////////////////////////////////////////\n";
    log << endl << endl;
    log << m;
    log << endl;
    log << endl;
    

    ////////////////////////////////////////////////////////////////////
    //
    //  Process the new message according to the message type code
    //
    ////////////////////////////////////////////////////////////////////
    type_of_message = m->getMsgType();
    switch (type_of_message){
      
      //----------------------------------------
      //     A D D    R E Q U E S T 
      //----------------------------------------
      case ADD_MESSAGE_TYPE:
	sql = m->getSQL();
	addr = m->getAddr();
	ap.id =            m->getPID();
	ap.clientID =      m->getCID();
	ap.port =          m->getPort();
	ap.period =        m->getPeriod();
	ap.numBaseTables = m->getNumBase();
	ap.clientAddr =    strdup( addr );
	ap.sqlString =     strdup( sql );
  	ap.dbname =        strdup( IDS_DB_NAME );
  	ap.deleted =       false;                 // this profile is not deleted yet
	delete[] sql;
	delete[] addr;

	//----------------------------------------
	//    A D D    R E Q U E S T
	//----------------------------------------
	if( ( st = addProfile( ap )) != OK){
	    delete m;
	    log << "\tStatus: "<<st<<endl<<endl;
	    freeProfile(ap);
	    return st;
	}
	break;
	
       //----------------------------------------
       //    D E L E T E    R E Q U E S T
       //----------------------------------------
       case DELETE_MESSAGE_TYPE:
	 if((st = deleteProfile(m->getPID(), m->getCID())) !=OK ){
	     delete m;
	     log << "\tStatus: "<<st<<endl<<endl;
	     return st;
	 }
	 break;      
	 
       //-----------------------------------------
       //    U P D A T E    R E Q U E S T 
       //----------------------------------------
       case UPDATE_MESSAGE_TYPE:
	   sql = m->getSQL();
	   if( (st = updateClientParameters( sql )) != OK ){
	       delete m;
	       delete[] sql;
	       log << "\tStatus: "<<st<<endl<<endl;
	       return st;
	   }
	   delete[] sql;
	   break;
	   
	//----------------------------------------
	//    I N V A L I D    M E S S A G E  
	//----------------------------------------
        default:
	    ERROR("Invalid message type");
	    return FAIL;
	    break;  
      
    }
    
    delete m;
    log << "\tStatus: "<<st<<endl<<endl;
    return OK;
}


//----------------------------------------------------------------------------------------
//  addProfile() - This method takes a generated profile (after badd query translation)
//                 and adds it to the profiles table in the main db.  It also adds it to 
//                 the timerTable multimap (shared in-mem struct used to store profiles 
//                 while smartPush runs).  If the profile exists, error is returned to caller
//  
//  ARGS:    ap - the profile to be added
//
//  RETURN:  error code { OK, PROFILE_EXISTS, ADD_FAILED }
//----------------------------------------------------------------------------------------
int 
idsReceiver::addProfile( AProfile ap )
{
    
    int         st, numTuples=0;
    ostrstream  os, os2;

    //--------------------------------------------------
    //  Check if profile exists before adding
    //--------------------------------------------------
    os << "SELECT * FROM PROFILES P WHERE P.ID="<<ap.id
       << " AND P.CID="<<ap.clientID<<ends;
    if( (st = _query->execute(os.str(), numTuples)) != OK){
	os.freeze(0);
	return (st);
    }
    if (numTuples != 0){
	WARN( "Profile id "<<ap.id<<" already exists for client "<<ap.clientID); 
	os.freeze(0);
	return ( PROFILE_EXISTS );
    }
    
    os.freeze(0); // unfreeze the stream
    

#ifdef DEGUG    // Debugging code only
    cout << "\n\nRECEIVER THREAD ADD: (sql) "<<ap.sqlString<<endl;
    cout << "IDS_RECEIVER THREAD ADD: (pid) "<<ap.id<<endl;
    cout << "IDS_RECEIVER THREAD ADD: (cid) "<<ap.clientID<<endl;
    cout << "IDS_RECEIVER THREAD ADD: (adr) "<<ap.clientAddr<<endl;
    cout << "IDS_RECEIVER THREAD ADD: (dnm) "<<ap.dbname<<endl;
    cout << "IDS_RECEIVER THREAD ADD: (per) "<<ap.period<<endl;
    //cout << "IDS_RECEIVER THREAD ADD: (tsl) "<<ap.ts_last.tv_sec<<endl;
    cout << "IDS_RECEIVER THREAD ADD: (por) "<<ap.port<<endl;
    cout << "IDS_RECEIVER THREAD ADD: (nbt) "<<ap.numBaseTables<<endl<<endl;
#endif DEBUG
    
    //--------------------------------------------------------------------------------
    //   Construct and execute an insert stmt for the profile table
    //   proID | cliID | cliAddr | SQL | dbname | period | TS | port | numTempTables|         
    //   change '0' to "NOW" if time adt is used
    //--------------------------------------------------------------------------------
    os2 << "INSERT INTO PROFILES VALUES ("<<ap.id<<", "<<ap.clientID<<", '"
	<< ap.clientAddr<<"', '"<<ap.sqlString<<"', '"<<IDS_DB_NAME<<"', "
	<< ap.period<<", 0 , "<<ap.port<<", "<<ap.numBaseTables<<")"<<ends;
    if( (st = _query->execute(os2.str(), numTuples)) != OK){
	os2.freeze(0);
	return (st);    
    }
    os2.freeze(0);

    if (numTuples != 1){
	ERROR("Insertion of profile id "<<ap.id<<" for client "<<ap.clientID<< " failed!"); 
	ERROR("Query: "<<os.str());
	return (ADD_FAILED);
    }

    //----------------------------------------------------------------------
    // Insert the profile to the in memory timer table
    //----------------------------------------------------------------------
    timespec t;
    t.tv_sec =  ap.period;    // 0=ts_last, ts_next is 0+p.period
    t.tv_nsec = 0;
    
    //---------------
    //  Lock TT
    //---------------
    timerTable.lock();
    timerTable.insert( timerTableValueType( t,ap ), true ); 
    timerTable.signal();

    //---------------
    //  Unlock  TT
    //---------------
    timerTable.unlock();
    
    return OK;
}


//----------------------------------------------------------------------------------------
//  deleteProfile() - This method deletes a profile from the main DB and the timer table.
//                    If the profile does not exist, then false is returned to caller
//  
//  ARGS:    pid - profile id
//           cid - client id
//
//  RETURN:  error code { OK, NO_SUCH_PROFILE }
//----------------------------------------------------------------------------------------
int 
idsReceiver::deleteProfile( int pid, int cid )
{
    
    int         st, i, numTuples;
    ostrstream  os;
    bool        foundProfile = false;
    
    //----------------------------------------
    //  Delete the profile from the DB
    //----------------------------------------
    os << "DELETE FROM PROFILES WHERE ID="<<pid
       << " AND CID="<< cid <<ends;       

INFO("DELETING, getting the RW lock");

    //----------------------------------------
    //  Obtain exclusive writer lock
    //----------------------------------------    
    RWL.writerLock();    
INFO("DELETING, got the RW lock");

    if( (st = _query->execute(os.str(), numTuples)) != OK) {
	RWL.writerUnlock();
	os.freeze(0);
	return (st); 
    }
    os.freeze(0);
    if (numTuples == 0){ 
	RWL.writerUnlock();
	return (OK); 
    }
    
    
    //----------------------------------------
    // Delete profile from the timer table
    //----------------------------------------
    // timerTable.lock();  // grab the mutex for the timerTable
    timerTablePtrType ix;
    
    if( timerTable.empty(true) ){
	RWL.writerUnlock();
	return (NO_SUCH_PROFILE);       
    }
    for( ix = timerTable.begin(true); ix != timerTable.end(true);  ix++){
	if((*ix).second.id == pid &&
	   (*ix).second.clientID == cid) {

INFO("DELETING, removing from timer table");	    

#ifdef DEBUG
//    cout << "Erasing TimerTable Element:"<<endl;
//	    cout << "----------------------------"<<endl;
//    
//    cout << "RT DELETE: (sql) "<<(*ix).second.sqlString<<endl;
//    cout << "RT DELETE: (pid) "<<(*ix).second.id<<endl;
//    cout << "RT DELETE: (cid) "<<(*ix).second.clientID<<endl;
//    cout << "RT DELETE: (adr) "<<(*ix).second.clientAddr<<endl;
//    cout << "RT DELETE: (dnm) "<<(*ix).second.dbname<<endl;
//    cout << "RT DELETE: (per) "<<(*ix).second.period<<endl;
//    cout << "RT DELETE: (tsl) "<<(*ix).second.ts_last.tv_sec<<endl;
//    cout << "RT DELETE: (por) "<<(*ix).second.port<<endl;
//    cout << "RT DELETE: (nbt) "<<(*ix).second.numBaseTables<<endl<<endl;
//    cout << "RT DELETE: (sql) "<<(*ix).second.sqlString<<endl<<endl;
#endif DEBUG

	    
	    // ----------------------------------------
	    // Found the element, delete it and break
	    // ----------------------------------------
	    timerTable.erase(ix,true);                    // remove entry
	    timerTable.signal();
	    foundProfile = true;
	    break;
	}
    }
    // timerTable.unlock();   // release the mutex for the timerTable

    if(! foundProfile ){
	ERROR("Profile not found in timer table");
    }

INFO("DELETING, releasing rw lock");	    
 
    RWL.writerUnlock();
    return OK;
}


//----------------------------------------------------------------------------------------
//  updateClientParameters() - Execute an update client parameters statement receieved in
//                             an update parameters message
//  
//  ARGS:    sqlstring - the SQL stmt to execute
//
//  RETURN:  error code - 
//----------------------------------------------------------------------------------------
int  
idsReceiver::updateClientParameters( char* sqlstring )
{
    int st, numTuples;
    
    if( (st = _query->execute(sqlstring, numTuples)) != OK ){
	return (st);
    } 
    if (numTuples == 0){
	WARN("No tuples affected by Update stmt");
    }
    
    INFO( "UPDATE CLI PARAMS SQL: "<< sqlstring);
    return OK;
}


//----------------------------------------------------------------------------------------
//  sendError() - Respond to an error condition.  Send the error code to the
//                client TCP sock, close the connection, and display a message
//
//  ARGS:    sock - the TCP socket to send an error to
//           errorcode - the error code to send
//
//  RETURN:  none
//----------------------------------------------------------------------------------------
void 
idsReceiver::sendError(TCPSock* sock, int errorcode)
{
    BEM.display(errorcode);
    sock->put(errorcode);  // pass the error code back to message sender
    sock->close();  
}
