
#include <ids_updater_thread.h>


//----------------------------------------------------------------------------------------
//  constructor() - Initialize the updater_thread.  Set up a query object and connect it
//                  to the IDS DB.  Set up the receiver socket for incoming requests
//  
//  ARGS:    none
//  RETURN:  none
//----------------------------------------------------------------------------------------
updaterThread::updaterThread()
{
    char * hostname;
    connection = NULL;
    cursor = NULL;
    int ret;
    
    hostname = getenv("PDHOST");
    if(!hostname){
	hostname = "localhost";
    }       
    
    connection = new ParaConnect(hostname, PARADISE_PORT);
    if( ! connection ){
	ERROR("Memory allocation failed, aborting");
	exit(1);
    }
    
    cursor = new ParaCursor(*connection);
    if( ! cursor ){
	ERROR("Memory allocation failed, aborting");
	exit(1);
    }
    
    ret = cursor->authenticate( TOR_USER, TOR_PASSWORD );
    if(ret != clNOERROR){
	ERROR("Failed to Authenticate user "<<TOR_USER);
	exit(1);
    }
}


//----------------------------------------------------------------------------------------
//  destructor() -
//
//  ARGS:    none
//  RETURN:  none
//----------------------------------------------------------------------------------------
updaterThread::~updaterThread()
{ }



//----------------------------------------------------------------------------------------
//  run() -   The run function for the ids_updater thread.  This thread updates profiles
//            at the ids site that have been added/modified locally.  It also executes 
//            parameter modification stmts at the ids site that have been executed locally
//
//  ARGS:    none
//  RETURN:  none
//----------------------------------------------------------------------------------------
int
updaterThread::run()
{
    
    char   queryStr[100];
    int    ret;
    
    
    for(;;) {
	
	// Code to ensure ctrl-C and ctrl-D get caught 
	signal( SIGINT, SIG_DFL );      // catch ctrl-C 
	signal( SIGQUIT, SIG_DFL );     // catch ctrl-D      
	
	sleep( PROFILE_UPDATE_TIME );
	
	INFO("Update thread checking for updates (time interval: "
	     <<PROFILE_UPDATE_TIME<< " seconds)");
	
	//-------------------------
	//   Open the Database
	//-------------------------
	sprintf(queryStr, "OPEN DB %s", CLIENT_DB_NAME);
	ret = cursor->declare( queryStr );
	if( ret != clNOERROR ){
	    ERROR("Could not bind query: "<< queryStr);
	    continue;
	}
	ret = cursor->exec();
	if(ret != clNOERROR){
	    ERROR("could not open db: "<<CLIENT_DB_NAME);
	    cursor->close();
	    continue;
	}
	
	//-----------------------------------------------------------------
	//   Scan Profiles and Parameters Tables for IDS updates
	//-----------------------------------------------------------------
	updateProfiles();
	updateParameters();
	
	
	//-------------------------
	//   Close the Database
	//-------------------------
	sprintf(queryStr, "CLOSE DB %s", CLIENT_DB_NAME);	    
	ret = cursor->declare( queryStr );
	if( ret != clNOERROR ){
	    ERROR("could not bind query: "<< queryStr);
	    continue;
	}
	ret = cursor->exec();
	if(ret != clNOERROR){
	    ERROR("could not close db: "<<CLIENT_DB_NAME);
	    cursor->close();
	}      
    }
}


//----------------------------------------------------------------------------------------
//  updateProfiles() -  scan the local profiles table and attempt to contact the IDS to 
//                      update any profile whose state != registered.  If a profile is 
//                      successfully updated at the IDS site, set the local state to 
//                      registered 
//
//  ARGS:    none
//  RETURN:  none
//----------------------------------------------------------------------------------------
void 
updaterThread::updateProfiles()
{
    
    ClientBaseAdt**            result;
    BaseAdtInfo::BaseAttrType  attrType;
    ClientInt4Adt *            intAdtPtr;
    int                        ret, st, err;
    int                        unregistered = 0;
    int                        registeredThisRun = 0;
    int                        unRegisteredThisRun = 0;
    int                        deletedThisRun = 0;
    int                        prof_state;
    bool                       connectStat;
    int                        prof_state_idx = 10;
    char                       queryStr[100];
    
    
    //----------------------------------------------------------------------
    //   Select all profiles that require updating at IDS site (adds, deletes)
    //   from our local profiles table
    //
    //   Registered = 0, all else require updating
    //----------------------------------------------------------------------
    sprintf(queryStr, "select * from MY_PROFILES where PROF_STATE > 0");
    
    // Open an updateable cursor over the result of above query (within a transaction)
    ret = cursor->declare(queryStr, true);
    if( ret != clNOERROR ){
	ERROR("could not bind query: "<< queryStr);
	cursor->close();
	return;
    }

    ret = cursor->open();
    if( ret != clNOERROR ){
	if(ret == clEND_TUPLES){
	    //INFO( "No profiles to update this time" );
	    cursor->close();
	    return;
	}
	else{
	    ERROR("could not open query: "<< queryStr);
	    cursor->close();
	    return;
	}
    }
    
    ret = cursor->fetch();
    assert ( ret != clEND_TUPLES );
    //if(ret == clEND_TUPLES){
    //	INFO( "No profiles to send to the IDS this time");	 
    //cursor->close();
    //return;
    //}
    
    int attrCount;
    ret = cursor->numAttrs( attrCount );
    if(ret != clNOERROR){
	ERROR( "Failed to get attrCount: ret="<<ret);	 
	cursor->close();
	return;
    }       
    assert( attrCount == 11 );
    
    ////////////////////////////////////////////////////////////////////////////////////////////
    //
    //   field      name                type        desc
    //   -----      -------------       --------    ------------------------------------
    //     1        pid                 Integer     This profile ID
    //     2        cid                 Integer     Our ID
    //     3        cliaddr             Integer     Our Address
    //     4        profile             String      The profile SQL string
    //     5        translation         String      The translated profile SQL string
    //     6        db_name             String      
    //     7        period              Integer     Period between profile executions 
    //     8        init_time           Integer     The initial Timestamp (normally 0)
    //     9        my_port             Integer     Our incoming port (for data from IDS)
    //     10       numTempTables       Integer     The number of tables in the from clause
    //     11       registered          boolean     Has this profile been registered at IDS
    //
    //////////////////////////////////////////////////////////////////////////////////////////
    
    while(!ret) {

	//----------------------------------------------------------------------
	// Open a socket to the IDS profile management receiver (for each update)
	//----------------------------------------------------------------------
	ids_sock = new TCPSock( IDS_HOSTNAME, IDS_RECEIVER_PORT, connectStat );
	INFO("Connecting to IDS: "<<IDS_HOSTNAME<<", "<<IDS_RECEIVER_PORT);
	if( connectStat == false ){
	    WARN("Connection to IDS failed, no IDS profile updates were made");
	    if(ids_sock)
		delete ids_sock;
	    cursor->close();
	    return;
	}
	
	unregistered++;
	ret = cursor->getTuple(result);
	AProfile pf;
	
	// retrieve each profile's fields and build a struct. 
	// Insert this struct into the in-memory timerTable
	for (int i=0; i<attrCount; i++)
	{
	    ClientBaseAdt *obj;
	    
	    ret = cursor->getObj(obj,attrType,i);
	    
	    switch (i) {
		
	    case 0:                       // profile id
		pf.id = 
		    (int)((ClientInt4Adt*)obj)->getIntVal();
		break;
	    case 1:                       // clientID
		pf.clientID = 
		    (int)((ClientInt4Adt*)obj)->getIntVal();
		break;
	    case 2:                       // the clientAddr
		pf.clientAddr = 
		    strdup((const char*)((ClientStringAdt*)obj)->getStrVal());
		break;
	    case 4:                       // the sqlString (profile)
		pf.sqlString = 
		    strdup((const char*)((ClientStringAdt*)obj)->getStrVal());
		break;
	    case 5:
		pf.dbname =                 // the main BADD dbname
		    strdup((const char*)((ClientStringAdt*)obj)->getStrVal());
		break;
	    case 6:                       // period for running this profile
		pf.period = (int)((ClientInt4Adt*)obj)->getIntVal();
		break;
	    case 7:                       // Time this profile was last run
	        pf.ts_last.tv_sec = 0;      //(int)((ClientInt4Adt*)obj)->getIntVal();
		pf.ts_last.tv_nsec=0;
		break;
	    case 8:                          // client smartPull port
		pf.port = (int)((ClientInt4Adt*)obj)->getIntVal();
		break;
	    case 9:                          // the number of temp tables used to ex. this prof.
		pf.numBaseTables = (int)((ClientInt4Adt*)obj)->getIntVal();
		break;
	    case 10:
		prof_state = (int)((ClientInt4Adt*)obj)->getIntVal();
		break;
	    case 3:  //translation, registered
		break;
	    default:
		assert(0);
		break;
	    }
	}
	
	//----------------------------------------------------------------------
	//   Send message to the IDS, if success, then update 
	//   registered field
	//----------------------------------------------------------------------
	assert( prof_state > 0 && prof_state < MAX_PROFILE_STATES); 
	if( prof_state == ADD_PROFILE ){
	    st = sendProfileAddMessage( pf );
	    
	    if( st == OK ){
		// update the prof_state field to OK
		//INFO( "IDS profile registration succeeded");
		
		registeredThisRun++;		
		
		ClientBaseAdt *obj=NULL;
		
		// get the prof_state field
		err = cursor->getObj( obj, attrType, prof_state_idx);
		if(err) { 
		    cursor->close();
		    delete ids_sock;
		    return;
		}
		
		// get the state object and update it's value
		intAdtPtr = (ClientInt4Adt *)result[prof_state_idx];
		intAdtPtr->setIntVal( (int)OK );
		
		int modIdxs[1];
		modIdxs[0] = prof_state_idx;
		
		/////////////////////
		//
		// Update prof_state
		//
		/////////////////////
		err = cursor->update( result, 1, modIdxs);
		
		if(err == clNOERROR){
		    INFO( "Local profile is registered at IDS");
		}
	    }else{
		ERROR( "Add message failed to register at IDS site: st=" <<st );
	    }
	}
	
	// Send a delete message
	else if( prof_state == DELETE_PROFILE ){
	    st = sendProfileDeleteMessage( pf.id, pf.clientID );
	    if( st == OK ){	      
		deletedThisRun++;
		err = cursor->remove();	
	
		if( err != clNOERROR){
		    ERROR( "Cursor REMOVE() failed!: "<<cem[err].message);
		}				
	    }else{
		ERROR( "Delete message failed to commit at IDS site: "<<st );
	    }
	}

	//----------------------------------------
	//  Close connection, get next tuple
	//----------------------------------------
	delete ids_sock;
	ids_sock = NULL;
	ret = cursor->fetch();	 
	
    }// end scan of non-updated profiles    
    
    
    if ( ret != clEND_TUPLES ) {
	ERROR( "Cursor Error: "<< cem[ret].message);
    }
    cursor->close();
    return;        
}


//----------------------------------------------------------------------------------------
//  updateParameters() - Update parameters by executing a SQL stmt at the IDS site.
//
//  ARGS:    none
//  RETURN:  none
//----------------------------------------------------------------------------------------
void 
updaterThread::updateParameters()
{
    
    ClientBaseAdt**            result;
    BaseAdtInfo::BaseAttrType  attrType;
    ClientInt4Adt *            intAdtPtr;
    int                        ret, st, err;
    int                        updateStmtsRun = 0;
    bool                       connectStat;
    char                       queryStr[100];
    char *                     updateString;
    int                        pid;
    int                        sqllen;
    int                        retVal;
    char *                     send_buf; 

    //----------------------------------------------------------------------
    //   Select all update stmts that need to be run at IDS site
    //   from our local update table
    //   Open an updateable cursor over the result so if update succeeds
    //   we delete this ids update from out table
    //----------------------------------------------------------------------
    sprintf(queryStr, "select * from IDS_UPDATES where STM_STATE > 0");
    ret = cursor->declare(queryStr, true);  
    if( ret != clNOERROR ) {
	ERROR("Unable to bind query: "<< queryStr);
	return;
    }
    
    ret = cursor->open();
    if( ret != clNOERROR ) {
	if( ret == clEND_TUPLES ){
	    return;
	}
	else{
	    ERROR("Unable to open query: " << queryStr);
	    cursor->close();
	    return;
	}
    }

    ret = cursor->fetch();             
    if( ret != clNOERROR ) {
	ERROR("Fetch failed");
	cursor->close();
	return;
    }
    assert( ret != clEND_TUPLES);

    
    int attrCount;
    ret = cursor->numAttrs( attrCount );
    if(ret != clNOERROR){
	ERROR( "Unable to get numAttr");
	cursor->close();
	return;
    }
    assert( attrCount == 3 );
    
    
    ////////////////////////////////////////////////////////////////////////////////////////////
    //
    //   field      name                type        desc
    //   -----      -------------       --------    ------------------------------------
    //     0        pid                 Int         The profile this parameter stmt is for
    //     1        update_stmt         String      The SQL string to execute at IDS
    //     2        stm_state           Int         The state of this update stmt
    //
    //////////////////////////////////////////////////////////////////////////////////////////
    
    //......................................................................
    // Get each tuple, send a param update message to the IDS
    // if message is executed at IDS, delete update string here
    //......................................................................
    while(!ret) {

	//-------------------------------------------------------
	// Open a socket to the IDS profile management receiver
	//-------------------------------------------------------
	ids_sock  = new TCPSock( IDS_HOSTNAME, IDS_RECEIVER_PORT, connectStat );
	if( connectStat == false ){
	    ERROR("Failed to connect to IDS, no IDS parameter updates were made");
	    if(ids_sock){
		delete ids_sock;
	    }
	    cursor->close();
	    return;
	}
    
	
	ret = cursor->getTuple(result);
	if( ret != clNOERROR ) {
	    ERROR("Unable to get tuple");
	    cursor->close();
	    delete ids_sock;
	    return;
	}
	
	//..................................
	//  Retrieve sql string from tuple
	//..................................
	ClientBaseAdt *obj;	    
	ret = cursor->getObj(obj,attrType,1);
	if( ret != clNOERROR ) {
	    ERROR("Unable to get object from tuple");
	    cursor->close();
	    delete ids_sock;
	    return;
	}	
	updateString = strdup((const char*)((ClientStringAdt*)obj)->getStrVal());

	ret = cursor->getObj(obj,attrType,0);
	if( ret != clNOERROR ) {
	    WARN("Unable to get tuple");
	    cursor->close();
	    delete ids_sock;
	    free(updateString);
	    return;
	}
	pid = (int)((ClientInt4Adt*)obj)->getIntVal();
	
	
	//-----------------------------------------------------------------------------
	//  Contact the IDS site and send this parameter update request, get err code
	//-----------------------------------------------------------------------------
	pm_message * m = new pm_message( UPDATE_MESSAGE_TYPE, 
					 CLIENT_ID, 
					 pid, 
					 strlen(updateString)+1, 
					 updateString );

	send_buf = new char[m->getMsgLength()];
	m->getBytes( send_buf );	
INFO("SENDING UPDATE MESSAGE: " << m);	
        ids_sock->put( m->getMsgLength() ); 
	retVal = ids_sock->put_bytes( send_buf, m->getMsgLength() );
	//delete send_buf;
INFO("SENT, GETTING RESPONSE NOW ..");	
	delete m;
	//-------------------------------
	// Get retVal from the IDS site
	//-------------------------------
	ids_sock->get( retVal );	
INFO("DONE, GOT: "<<retVal);
	
	if( retVal != OK ){
	    WARN( "IDS Profile registration failed: code"<< UPDATE_PARAM);
	}	
	
	//----------------------------------------------------
	//  Delete this update stmt if the IDS executed it OK
	//----------------------------------------------------
	else{
	    
	    err = cursor->remove();
	    if(err == clNOERROR){
	    INFO( "Local parameter update stmt has been deleted!");
	    }
	    else{
	    	ERROR( "Failed to remove parameter update stmt from ids_updates table");
	    }	      
	}
	
	//----------------------------------------
	//  Close connection, get next tuple
	//----------------------------------------
	delete ids_sock;
	ids_sock = NULL;
	ret = cursor->fetch();
    }    
    
//INFO( "Scan of update stmts done:");
    cursor->close();
    return;
}


//----------------------------------------------------------------------------------------
//  sendProfileAddMessage() - Sends a profile add message to the IDS site  
//
//  ARGS:    ap - the profile to register at the main IDS site
//  RETURN:  error code - { FAIL, err code from IDS site }
//----------------------------------------------------------------------------------------
int 
updaterThread::sendProfileAddMessage( AProfile ap )
{
    
    int       sqllen, addlen, retVal;
    char *    send_buf;
    
    //-----------------------------------------------------
    //   Create a message object to send to the IDS
    //-----------------------------------------------------
    pm_message * m = new pm_message( ADD_MESSAGE_TYPE, 
				     CLIENT_ID, 
				     ap.id,
				     ap.port,
				     ap.period,
				     ap.numBaseTables,
				     strlen(ap.clientAddr)+1,
				     strlen(ap.sqlString)+1, 
				     ap.clientAddr,
				     ap.sqlString);   
    
    send_buf = new char[m->getMsgLength()];
    m->getBytes( send_buf );	
INFO("SENDING ADD MESSAGE: "<<m);	
    ids_sock->put( m->getMsgLength() ); 
    retVal = ids_sock->put_bytes( send_buf, m->getMsgLength() );
INFO("SENT, GETTING RESPONSE NOW ..");	
    
    //-------------------------------
    // Get retVal from the IDS site
    //-------------------------------
    ids_sock->get( retVal );	
    if( retVal != OK ){
	WARN( "IDS Profile registration failed: code"<< UPDATE_PARAM);
    }	
INFO("DONE: returning err code: "<< retVal);
 
//delete[] send_buf; 
    delete m;
    return retVal;	
}    


//----------------------------------------------------------------------------------------
//  sendProfileDeleteMessage() - Send a drofile delete message to the IDS and return 
//                               the error code received from the IDS to the caller
//
//  ARGS:    pid - the profile id to delete
//           cid - the client who owns the profile
//
//  RETURN:  err code { FAIL, IDS return code }
//----------------------------------------------------------------------------------------
int 
updaterThread::sendProfileDeleteMessage( int pid, int cid )
{
    int retVal = 0;
    char * send_buf;
    
    pm_message * m = new pm_message( DELETE_MESSAGE_TYPE, cid, pid );
    send_buf = new char[m->getMsgLength()];

    m->getBytes( send_buf );	
INFO("SENDING DELETE MESSAGE: "<< m);

    ids_sock->put( m->getMsgLength() ); 
    retVal = ids_sock->put_bytes( send_buf, m->getMsgLength() );

INFO("SENT, GETTING RESPONSE NOW ..");	
    
    //-------------------------------
    // Get retVal from the IDS site
    //-------------------------------
    ids_sock->get( retVal );	
    if( retVal != OK ){
	WARN( "IDS Profile registration failed: code"<< UPDATE_PARAM);
    }	

INFO("DONE: returning err code: "<< retVal);

    delete m;
    return retVal;
}
