
#include <profile_thread.h>

//----------------------------------------------------------------------------------------
//  constructor() -  Initialize this profile thread with the data needed to execute 
//                   this profile.
//
//  ARGS:    profile - the profile this thread will execute
//           retCode - status code indicating if profile thread was created OK
//  RETURN:  retCode (by ref)
//----------------------------------------------------------------------------------------
profileThread::profileThread(AProfile profile , bool & retcode)
{
    ostrstream  tempQueries, finalQueries;
    char *      ptr;
    int         qryNum, position, i;
    queryMapPtrType qMapPtr;

    assert(profile.sqlString != NULL);
    assert(profile.clientAddr != NULL);
    assert(profile.dbname != NULL);

    retcode = true;  // assume we will init OK
    
    _profile.id = profile.id;
    _profile.clientID = profile.clientID;
    _profile.period = profile.period;
    _profile.ts_last = profile.ts_last;
    _profile.port = profile.port;
    _profile.numBaseTables = profile.numBaseTables;

    _profile.clientAddr = strdup(profile.clientAddr);
    _profile.sqlString = strdup(profile.sqlString);
    _profile.dbname = strdup(profile.dbname);
    
    numTempQueries  = profile.numBaseTables + 1;
    numFinalQueries = profile.numBaseTables;
    numTotalQueries = numTempQueries+numFinalQueries;
    
    // allocate an array of strings to hold this profiles queries
    // queryStrings = new char[numTotalQueries][DEFAULT_LEN];
    
    char * tmpStr = new char[DEFAULT_LEN];
    if(!tmpStr){
	retcode = false;
	threadQueue->Lock();
	threadQueue->Put(this);
	threadQueue->Unlock();
	return;
    }
    bzero(tmpStr, DEFAULT_LEN);

    // split profile SQL string into setup and final querie strings
    if( ! (split_string( _profile.sqlString, tempQueries, finalQueries,'$')) ){
	WARN( "No temp queries found");
    }	
    tempQueries << ends;
    finalQueries << ends;
    
    //----------------------------------------------------------------------
    //   load each temp (setup) query into the queries array
    //----------------------------------------------------------------------
    qryNum=0;
    position=0;
    ptr = tempQueries.str();
    while (*ptr){
	if(*ptr==';') {
	    
	    // End query string with null, and insert to map
	    tmpStr[position] = '\0';
	    queryStringsMap.insert( queryMapValueType( qryNum, tmpStr ) );
	    tmpStr = NULL;


	    // Allocate a new query string array
	    tmpStr = new char[DEFAULT_LEN];
	    if(!tmpStr){
		deleteQueryStrings();
		retcode = false;
		threadQueue->Lock();
		threadQueue->Put(this);
		threadQueue->Unlock();
		return;
	    }
	    bzero(tmpStr, DEFAULT_LEN);

	    qryNum++;
	    position=0;
	}
	else{
	    tmpStr[position] = *ptr;
	    position++;
	}
	ptr++;
    }
    
    //----------------------------------------------------------------------
    //   load each final (dissemination) query into the queries array
    //----------------------------------------------------------------------
    position = 0;
    ptr = finalQueries.str();
    while (*ptr){
	if(*ptr==';') {

	    // End query string with null, and insert to map
	    tmpStr[position] = '\0';
	    queryStringsMap.insert( queryMapValueType( qryNum, tmpStr ) );
	    tmpStr = NULL;
	    
	    qryNum++;
	    position=0;
	    if(qryNum == numTotalQueries){
		break;
	    }

	    // Allocate a new query string array
	    tmpStr = new char[DEFAULT_LEN];
	    if(!tmpStr){
		deleteQueryStrings();
		retcode = false;
		threadQueue->Lock();
		threadQueue->Put(this);
		threadQueue->Unlock();
		return;
	    }
	    bzero(tmpStr, DEFAULT_LEN);

	}
	else{
	    tmpStr[position] = *ptr;
	    position++;
	}
	ptr++;
    }

    tempQueries.freeze(0);
    finalQueries.freeze(0);
    
    //------------------------------------------------------------
    //  Init a query object for this profile thread
    //------------------------------------------------------------
    query = NULL;
    query = new SQLQuery( IDS_DB_NAME );
    
    if( ! query ){
	ERROR("Failed to allocate query object");
	retcode = false;
	deleteQueryStrings();
	threadQueue->Lock();
	threadQueue->Put(this);
	threadQueue->Unlock();
	return;
    }
    
    else if( query->open_db() != OK ) {
	ERROR("Failed to open DB in profile thread constructor");
	retcode = false;
	deleteQueryStrings();
	threadQueue->Lock();
	threadQueue->Put(this);
	threadQueue->Unlock();
	return;
    }
}


//----------------------------------------------------------------------------------------
//  destructor() - 
//
//  ARGS:    none
//  RETURN:  none
//----------------------------------------------------------------------------------------
profileThread::~profileThread() 
{ 
    /* all is done before thread termination */ 
}


//----------------------------------------------------------------------------------------
//  createTempTables()  -  Creates the temp tables needed by a profile while it runs.
//                         the number of temp tables needed is stored with the profile
//  ARGS:    none
//  RETURN:  error code { OK, DB_OPEN_ERROR }
//----------------------------------------------------------------------------------------
int 
profileThread::createTempTables()
{
    
    char  tmp[256];
    int   ret = 0, numTuples=0;
    
    //--------------------------------------------------
    //  construct the attrList for the first tmp table 
    //  (this will depend on the number base tables)
    //--------------------------------------------------
    ostrstream attrList;
    for(int i=1; i<=_profile.numBaseTables; i++)
	attrList << "B"<<i<<"TS INTEGER, B"<<i<<"ID INTEGER, ";	
    attrList << "ISOLD BOOL"<<ends;
    
    //////////////////////////////////////////////////////////////////////////////
    //  create first temp table table:  name/format will be:
    //  <clientID_profileID_R( B1TS Int, B1ID Int ... BnTS Int BnID Int, .. old Bool)>
    //
    // where: 
    //    B1TS = timestamp field from base table 1
    //    B1ID = ID field from base table 1
    //     :
    //    BnTS = timestamp field from base table n
    //    BnID = ID field from base table n
    //    old = boolAnd of old values.  
    //         If all joining tuple are old, then true
    //         otherwise, false if any joining tuple is new
    /////////////////////////////////////////////////////////////////////////////// 
    sprintf(tmp, "CREATE TABLE T%d_%d_R(%s)",_profile.clientID, 
	    _profile.id, attrList.str());
    
    ret = query->execute(tmp, numTuples);
    if( ret ) {
	WARN("Create table stmt failed:"<<tmp);;
    }
    
    ///////////////////////////////////////////////////////////////////////////
    // Create ONE temp table for each base table in this profile 
    // to use as a second temp.  Each final query will join a base 
    // table with one of these temp tables.
    // table name/form for table n will be:
    // 
    //   <clientID_profileID_INT(BnID Integer, minOfOld Bool)>
    //
    ///////////////////////////////////////////////////////////////////////////
    for (int i=1; i<= _profile.numBaseTables; i++){
	sprintf(tmp, "CREATE TABLE T%d_%d_%d(B%dID INTEGER, MINOFOLD BOOL)",
		_profile.clientID, _profile.id, i, i);
	ret = query->execute(tmp, numTuples);
	if( ret ) {
	    WARN("Create table stmt failed: "<<tmp);
	}
    }
    
    attrList.freeze(0);
    return OK;
}



//----------------------------------------------------------------------------------------
//  dropTempTables()  -  Drops the temp tables used by this profile
//
//  ARGS:    none
//  RETURN:  error code { OK,  DB_OPEN_ERROR } 
//----------------------------------------------------------------------------------------
int 
profileThread::dropTempTables()
{
    
    char  tmp[150];
    int   ret=0, numTuples=0;


    //----------------------------------------
    //   Drop the first temp table
    //----------------------------------------
    sprintf(tmp, "DROP TABLE T%d_%d_R",_profile.clientID, _profile.id);	
    ret = query->execute(tmp, numTuples);
    if(ret){
	WARN( "Profile thread drop table stmt 1 failed (non-existent table?)");
    }
    
    //----------------------------------------
    //   Drop the remaining temp tables
    //----------------------------------------
    for (int i=1; i<= _profile.numBaseTables; i++){
	sprintf(tmp, "DROP TABLE T%d_%d_%d",_profile.clientID, _profile.id, i);       
	ret = query->execute(tmp, numTuples);
	if( ret ) {
	    WARN("Profile thread drop table stmt 2 failed (non-existent table?), err: " << ret);
	}
    }
    
    return OK;
}


//----------------------------------------------------------------------------------------
//  run()  -  Execute the profile.  This is the "main" function for the profile thread.
//               1. Run the profile
//               2. Send the profile if there is a result
//               3. Update the in memory and database timerTable and PROFILES table            
//  ARGS:    none
//
//  RETURN:  error code
//----------------------------------------------------------------------------------------
int 
profileThread::run()
{
    int ret, i;
    
    // Create the temp tables needed for this profile
    if( createTempTables() != OK ){
	ERROR("Error occured while creating profiles temp tables");
    }
    
    // Run profile and generate a tar file to send to client
    ret = run_profile();
    
    //------------------------------------------------------------
    //   Send the tar file, if send fails do not update the db 
    //   timer table, but do update in mem timer table so that
    //   the profile will be run again.  The normal case for this
    //   occuring will be when the IDS computes a result but is
    //   unable to connect to the client.  In this case the data
    //   that should have been sent to the client will never get
    //   sent unless we do not update the DB ts field
    //------------------------------------------------------------
    if( ret == OK ){	
	
	//-------------------------------------------------------------
	//  Send data and Update the db PROFILES table if data sends
	//-------------------------------------------------------------
	if( send_profile() == OK ){ 
	    timespec ttime;
	    getTime( ttime );
	    //INFO("Updating DB profile TS to " << ttime.tv_sec);
	    query->run_update( _profile.id, _profile.clientID, ttime.tv_sec);
	}	
	
	else{
	    WARN("Failed to send profile "<<  _profile.id << 
		 " result to client: "<< _profile.clientID );
	    BEM.display( ret );
	}
    }
    else if( ret == NO_RESULT ){
	INFO( "No result for execution of profile "
	      <<  _profile.id <<", client "<<  _profile.clientID<< " this time");
    }
    
    
    //----------------------------------------
    //   Cleanup and terminate this thread
    //----------------------------------------    
    update_timerTable();
    dropTempTables();
    query->close_db();
    
    delete query;
    deleteQueryStrings();
    threadQueue->Lock();
    threadQueue->Put(this);
    threadQueue->Unlock();
    return OK;
}


//----------------------------------------------------------------------------------------
//  deleteQueryStrings()  -  Delallocate all query strings from the stl queryStringsMap map 
//
//  ARGS:    none
//  RETURN:  none
//----------------------------------------------------------------------------------------
void 
profileThread::deleteQueryStrings()
{
    freeProfile( _profile );

    queryMapPtrType qMapPtr;
    for(qMapPtr = queryStringsMap.begin(); qMapPtr != queryStringsMap.end(); qMapPtr++){
	if((*qMapPtr).second){
	    delete[] (*qMapPtr).second;

	}
	queryStringsMap.erase(qMapPtr);
    }
}


//----------------------------------------------------------------------------------------
//  run_profile()  -  Run the profile.  Status is returned, and may be one of the following:
//
//     OK             Everything is ok, and a tar file to send to client was created 
//     NO_RESULT      Everyting is OK, but there is nothing to send to client, no tar file 
//     TAR_ERROR      Execution of tar command failed 
//     CURSOR_ERROR   A cursor error occured, no tar file generated
//  
//  ARGS:    none
//  RETURN:  error code as described above { OK, NO_RESULT, TAR_ERROR, CURSOR_ERROR }
//----------------------------------------------------------------------------------------
int 
profileThread::run_profile()
{
    
    char   prof_dirname[100], tmp[DEFAULT_LEN]; //tmp2[DEFAULT_LEN]	;	
    char   full_path[100];
    int    ret, i, numTuples=0;
    int    cliTableNum=1;
    int    numFiles=0;
    
    //----------------------------------------------------------------------
    // construct the dirname and make a new dir <clientID.profileID.dir>
    //----------------------------------------------------------------------
    sprintf(prof_dirname,"%s%d.%d.dir", PUSH_DATA_PATH,_profile.clientID, _profile.id);
    sprintf(full_path, "%s/%s", PUSH_BASE_PATH, prof_dirname);
    ret = mkdir(full_path, S_IRWXU | S_IRWXO);
    
    //INFO("tried to make dir: "<<full_path);
    
    //--------------------------------------------------
    // if this dir exists for some reason, clean it
    //--------------------------------------------------
    if( ret == -1 ){
	//INFO("dir: "<<full_path<<" exists, cleaning it");
	clean_directory( full_path );
    }
    
    //-------------------------------------------------------
    // Execute the temp queries that set up the temp tables 
    // for the final dissemination queries
    //-------------------------------------------------------
    for(i=0; i < numTempQueries; i++){
	if( (ret = query->execute(queryStringsMap[i], numTuples)) != OK ){
	    WARN("Error executing setup query: " << queryStringsMap[i] );
	}
	//INFO("Executing setup query: "<< i << queryStringsMap[i] );
	//INFO("NumTuples: "<< numTuples << ", ret code: "<<ret );
    }
    
    
    //////////////////////////////////////////////////////////////////////////////
    //
    //  The remaining queries are FINAL queries that actually disseminate tuples 
    //  from the IDS base tables to the client. These queries use the execute
    //  and dump method rather than the execute.  A SINGLE "schema file" is created.
    //  This file will contain the SQL statements to reproduce the tables
    //  and insert the data on the pull side.  For each base table in the profile
    //  a data file is created for loading at the pull site.  In addition, if a 
    //  field contains data that must be written to a file (Text adt, Raster Adt..)
    //  a unique named binary file is created containing the object to be loaded
    //  on the pull side
    //
    //////////////////////////////////////////////////////////////////////////////
    for(i=numTempQueries; i<numTotalQueries; i++){
	
	// construct a tmp filename <dirname.cliID.profID.cliTableNum.out>
	// this will be the unique filename for each final query result
	sprintf(tmp,"%s/%d.%d.%d.out",full_path, _profile.clientID,_profile.id, cliTableNum);	
	
	//INFO("Executing final query: "<< i << queryStringsMap[i] );
	ret = query->execute_and_dump(queryStringsMap[i],prof_dirname, 
				      _profile.numBaseTables, 
				      cliTableNum, 
				      _profile.clientID, 
				      _profile.id );
	
	//INFO("DISSEMINATION QUERY: "<<queryStrings[i]);
	//BEM.display( ret );
	
	//INFO("Executing final query: "<< i << queryStringsMap[i] );
	//INFO("NumTuples: "<< numTuples << ", ret code: "<<ret );

	// Increment the number of files
	if( ret == OK ){
	    numFiles++;
	}
	
	else if( ret < 0 ) {
	    ERROR("Error while executing final query");
	    clean_directory ( full_path );
	    rmdir( full_path );
	    return( CURSOR_ERROR );
	}
	
	cliTableNum++;
    }// end final queries dump
    
    
    //--------------------------------------------------  
    //  Tar the directory <dirname> that contains the 
    //  data generated from running this profile
    //--------------------------------------------------
    if( numFiles == 0 ){
	clean_directory ( full_path );
	rmdir ( full_path );
	return NO_RESULT;
    }
    else{
	fflush(stdout);
	ostrstream tar_os;

	tar_os << "tar cfv " << full_path  << ".tar " << prof_dirname <<ends;
	
#ifdef PUR_TEST	  
	ret = 0;
#else
	ret = system(tar_os.str());
	ret = ret >> 8;
#endif
	tar_os.freeze(0);	  
	
	if(ret != 0){
	    return (TAR_ERROR);
	}
	clean_directory ( full_path );
	rmdir ( full_path );
	return OK;
    }
}



//----------------------------------------------------------------------------------------
//  send_profile()  -  Send the profile to the client.  
//
//     1. Open a TCP socket connection to the client
//     2. Send the profileID
//     3. Send the clientID
//     4. Send the tarred file size
//     5. Send the tarred file itself in BLOCK_SIZE chunks
//     6. Close connection
//     7. Return one of the following error codes:
//
//     OK                    Everything is ok, tar file was sent to client 
//     FAILED_TO_CONNECT     Could not establish connection
//     SEND_PROFILE_FAILED   Failed while sending data 
//
//  ARGS:    none
//  RETURN:  error code as described above  { OK, FAILED_TO_CONNECT, SEND_PROFILE_FAILED }
//----------------------------------------------------------------------------------------
int 
profileThread::send_profile()
{
    int            status;
    struct stat    fstatus;
    int            file_size;
    char           prof_dirname[100];
    ifstream       tarfile;
    char           dummy_tar[100];
    bool           con_stat;
    char *         tarName;
    

    //----------------------------------------
    //  Open a TCP socket to client
    //----------------------------------------
    TCPSock sock(_profile.clientAddr, _profile.port, con_stat);
    if( con_stat == false ){ 
	return (FAILED_TO_CONNECT);
    }
    
    //----------------------------------------
    //  Open the tar file to send
    //----------------------------------------
    sprintf(dummy_tar, "%s%s", PUSH_BASE_PATH, "/DATA/junk.tar");   // dummy tar file for purify version
    sprintf(prof_dirname,"%s%d.%d.dir.tar",PUSH_DATA_PATH,
	    _profile.clientID,_profile.id);
    
#ifdef PUR_TEST
    tarName = dummy_tar;
#else
    tarName = prof_dirname;
#endif
    tarfile.open(tarName);      
    stat(tarName, &fstatus);      
    file_size = fstatus.st_size;
    
    
    //-------------------------------------------------------
    //  Send the pid, cid, file size, and tar file
    //-------------------------------------------------------
    if( sock.put( _profile.id ) == false ){
	WARN("Could not send profile ID");
	return (SEND_PROFILE_FAILED);
    }
    if( sock.put( _profile.clientID ) == false ){
	WARN("Could not send client ID");
	return (SEND_PROFILE_FAILED);
    }
    if( sock.put(file_size) ==  false ){
	WARN("Could not send the tar file size");
	return (SEND_PROFILE_FAILED);
    }
    
    // Send the tar file in chunks of BLOCK_SIZE 
    int nblocks, nbytes;
    char *buf;
    buf = new char[BLOCK_SIZE];
    
    nblocks = (int)file_size/(int)BLOCK_SIZE;        // number of blocks 
    nbytes = (int)file_size%(int)BLOCK_SIZE;          // bytes remaining 
    
    while(nblocks > 0) {
	memset(buf,0,BLOCK_SIZE);
	tarfile.read(buf,BLOCK_SIZE);
	status = sock.put_bytes(buf,BLOCK_SIZE);
	if(status != BLOCK_SIZE){
	    ERROR(" Error in sending file ");
	    tarfile.close();
	    sock.close();
	    delete [] buf;
	    return (SEND_PROFILE_FAILED);
	}
	nblocks--;
    }
    memset(buf,0,nbytes);
    tarfile.read(buf,nbytes);
    status = sock.put_bytes(buf,nbytes);
    if(status != nbytes){
	ERROR(" Error in sending file ");
	tarfile.close();
	sock.close();
	delete [] buf;
	return (SEND_PROFILE_FAILED);
    }
    
    tarfile.close();
    sock.close();
    delete [] buf;
    INFO("Profile result of "<< file_size<<" bytes sent to :"<<_profile.clientAddr <<", "<< _profile.port);
    return (OK);
}


//----------------------------------------------------------------------------------------
//  update_timerTable()  -  Update the timer table.  
//
//     1. Set TS (time last run) to "NOW" for profile entry in db PROFILES table
//     2. Add period to current time and reinsert this profile into the in meme timerTable
//
//  ARGS:    none
//
//  RETURN:  error code { OK, UPDATE_FAILED }
//----------------------------------------------------------------------------------------
int 
profileThread::update_timerTable()
{
    
    int status;
    timespec ttime;
    
    // Update the timerTable with the key value 'timeofday + profile's period', 
    // and a copy of this profile generate the "next run time" key for this profile
    getTime( ttime );
    
    // create a copy of this profile
    AProfile pf;
    pf.id = _profile.id;
    pf.clientID = _profile.clientID;
    pf.clientAddr = strdup(_profile.clientAddr);
    pf.sqlString = strdup(_profile.sqlString);
    pf.dbname = strdup(_profile.dbname);
    pf.period = _profile.period;
    pf.ts_last.tv_sec = ttime.tv_sec;
    pf.port = _profile.port;
    pf.numBaseTables = _profile.numBaseTables;
    pf.deleted = false;
    
    timerTable.lock();  // grab the mutex for the timerTable
    
    ttime.tv_sec += _profile.period;
    
    // get two iterators for the timerTable
    timerTablePtrType ix;
    timerTablePtrType iy;
    
    bool empty;
    
    empty = timerTable.empty(true);
    
    ix = timerTable.begin(true);
    timerTable.insert( timerTableValueType( ttime,pf ), true );    
    iy = timerTable.begin(true);
    
    
    //timerTable.dump(12345);
    
    if( empty ) { 
	// this is the first and only element in the timerTable, SIGNAL
	//INFO( "SIGNALING: TimerTable was empty" );
	
	//Release this profile threads reader lock
	timerTable.signal();
	RWL.readerUnlock();
	timerTable.unlock();
	return OK;
    }	
    
    INFO("Cur number of profiles in timerTable: "<<timerTable.size(true));
    
    if( (*iy).second.id == pf.id &&
        (*iy).second.clientID == pf.clientID) {
	
	// This profile is at the front of the timer table again
	// Signal the timer thread which is currently waiting to spawn some other profile
	// (this thread is pre-empting the profile that would be spawned next)
	//INFO( "SIGNALING: the cur profile was inserted to the front of the list");
	
	//Release this profile threads reader lock
	timerTable.signal();
	
	RWL.readerUnlock();
	timerTable.unlock();
	return OK;
    }	
    
    //(ttime.tv_sec);	
    RWL.readerUnlock();
    timerTable.unlock();   // release the mutex for the timerTable
    return OK;
}


//----------------------------------------------------------------------------------------
//  clear_directory()  -  deletes all files from a profiles working directory
//
//  ARGS:    dirname - the dir to clean
//  RETURN:  none
//----------------------------------------------------------------------------------------
void 
profileThread::clean_directory(const char* dirname)
{
       int   st;
       char  tmp[100];
       
       DIR * dir = NULL;
       struct dirent *dirent;
       
       dir = opendir( dirname );
       if( dir ){
	 while ( (dirent = readdir( dir )) != NULL ){
	   if( (strcmp(dirent->d_name, ".") !=0 ) &&
	       (strcmp(dirent->d_name, "..") !=0 )){
	     sprintf(tmp, "%s/%s", dirname, dirent->d_name );
	     unlink(tmp);
	   }
	 }
	 (void)closedir( dir );
	 //delete dir;
       }
}




	// Update the db PPROFILES table
 	//timeval t;
 	//gettimeofday(&t,NULL);
 	//ttime.tv_sec=t.tv_sec;
 	//ttime.tv_nsec=0;	
	//status = query->run_update( _profile.id, _profile.clientID, ttime.tv_sec);   //was ttime	











