
#include <pull_receiver.h>


//------------------------------------------------------------------------------------------
//  constructor()  -  Initialize the receiver_thread.  Set up a query object and connect it
//                    to the smartPull DB.  Set up the receiver socket for incoming requests
//  ARGS:     none
//  RETURNS:  none
//------------------------------------------------------------------------------------------
pullReceiver::pullReceiver()
{
       bool con_st;
       _socket = NULL;
       _query = NULL;

       _socket = new TCPSock( PULL_RECEIVER_PORT, con_st );
       _query = new SQLQuery ( CLIENT_DB_NAME, PARADISE_PORT );       
 
       assert( con_st );
       assert( _socket != NULL );
       assert( _query != NULL );

       if( _socket == NULL || _query == NULL ){
            perror( "Failed to create socket or query object for receiver thread" );
	    exit(1);
       }
       if(con_st == false){
            perror("Failed to create socket for receiver thread");
	    exit(1);
       }
}


//------------------------------------------------------------------------------------------
//  destructor()  -  Close and delete the socket.
//
//  ARGS:     none
//  RETURNS:  none
//------------------------------------------------------------------------------------------
pullReceiver::~pullReceiver()
{
       _socket->close();
       delete _socket;	
       end();
}


//------------------------------------------------------------------------------------------
//  run()  -  The main run thread for pull receiver thread
//
//  ARGS:     none
//  RETURNS:  none
//------------------------------------------------------------------------------------------
int 
pullReceiver::run(){

       int status, i;

       INFO( "\nPULL_RECEIVER_PORT: " << PULL_RECEIVER_PORT );
       INFO( "IDS_RECEIVER_PORT:  " << IDS_RECEIVER_PORT );
       INFO( "PULL_PARADISE_PORT: " << PARADISE_PORT<<endl);

       //---------------------------------------------------------------------
       //   Listen for profile mngmt requests and process them when received
       //---------------------------------------------------------------------
       for(;;) {

            // Code to ensure ctrl-C and ctrl-D get caught 
            signal(SIGINT, SIG_DFL);      // catch ctrl-C 
            signal(SIGQUIT, SIG_DFL);     // catch ctrl-D      
       
	    //----------------------------------
            //  Got a request -- accept it 
	    //----------------------------------
            _sendersSocket = _socket->accept();

	    INFO( "PULL RECIEVER THREAD: Got Profile Management Request" );

	    //--------------------
	    // process message 
	    //--------------------
	    if( _sendersSocket ){

	         // Open the DB and send
	         if( _query->open_db() == OK){
		    status = process_msg();
		    if(status != OK){
		         WARN("Error in get: error code passed back to message sender" );
		    }
		    _query->close_db();
	       }
	       else{
		    ERROR("Unable to open the DB");
		    sendError(_sendersSocket, DB_OPEN_ERROR);
	       }
	    }      

	    INFO( "RECIEVER THREAD: Request Processed, listening again" );
       }
}


//------------------------------------------------------------------------------------------
//  process_msg()  -  Process a profile management request message and responds accordingly
//
//  ARGS:     none
//  RETURNS:  error code  { FAIL, OK }
//------------------------------------------------------------------------------------------
int 
pullReceiver::process_msg()
{

    char *     address;
    char *     querystring;
    char *     updatestring;
    int        addlen = 0, sqllen = 0, update_len = 0;
    char       buf[1024];
    int        messageType;
    time_t     my_time = time(NULL);
    ofstream   log(PULL_RECEIVER_LOG,ios::app);
    AProfile   ap;
    bool       status; 
    int        st;
    
    
    /////////////////////////////////////////////////////////////////////////////////
    //
    //  All incoming messages will begin with
    // 
    //  1) The message type code
    //      1: add profile
    //      2: delete profile
    //      3:  (defunct)
    //      4: parameter action (add/delete) request
    //      5: parameter view request
    // 
    //  2) The profile id           (int) 
    //  3) The client id            (int) 
    //
    //  Other parts of the message will then be received based on the message type 
    //  code. 
    //
    /////////////////////////////////////////////////////////////////////////////////
    
    // Receive the message type code 
    if( _sendersSocket->get( messageType ) == false ){
	return FAIL;
    }
    // Receive the profile ID 
    if( _sendersSocket->get( ap.id ) == false  ){	 
	return FAIL;
    }
    // Receive the Client ID 
    if( _sendersSocket->get( ap.clientID ) == false  ){
	return FAIL;
    }
    
    //INFO("Pull Got Profile Managment Request");
    
    //----------------------------------------------------------------------
    //  Write this profile transaction to a log for debugging
    //----------------------------------------------------------------------
    log << "<--PULL PROFILE MANAGER RECEIVED REQUEST:"<<endl;
    log << "........................................................." <<endl;
    log << "  transaction time\t" << ctime(&my_time);
    log << "  code: \t\t";
    if(messageType == 1) log << "Profile ADD request"<<endl;
    if(messageType == 2) log << "Profile DELETE request"<<endl;
    if(messageType == 4) log << "Parameter ACTION request"<<endl;
    if(messageType == 5) log << "Parameter VIEW request"<<endl;
    
    log  <<"  pid: \t\t\t"    <<ap.id<<endl
	 <<"  cid: \t\t\t"    <<ap.clientID<<endl;
    
    
    switch(messageType){
	
	// This is a profile add or update
    case 1:
	/* Get the client port */
	if( _sendersSocket->get(ap.port) == false  ) return FAIL;  
	
	/* Get the client period */
	if( _sendersSocket->get(ap.period) == false  ) return FAIL;
	
	/* Get the client address length */
	if( _sendersSocket->get(addlen) == false  ) return FAIL;
	
	/* Get the client sql string length */
	if( _sendersSocket->get(sqllen) == false  ) return FAIL;
	
	address      = new char[addlen+1];
	querystring  = new char[sqllen+1];
	
	/* Get the client address */
	memset(buf,0,BLOCK_SIZE);
	if( _sendersSocket->get_bytes( buf, addlen ) != addlen) return FAIL;
	memcpy(address, buf, addlen);
	address[addlen]='\0';
	ap.clientAddr = strdup(address);
	delete[] address;
	
	/* Get the client sqlstring */
	memset(buf,0,BLOCK_SIZE);
	if( _sendersSocket->get_bytes( buf, sqllen ) != sqllen) return FAIL;
	memcpy(querystring, buf, sqllen);
	querystring[sqllen]='\0';
	
	log << "  Port: \t\t"         << ap.port             << endl
	    << "  Period: \t\t"       << ap.period           << endl
	    << "  Address: \t\t"      << ap.clientAddr       << endl
	    << "  Query: \t\t"        << querystring         << endl;    
	break;
	
	// Delete request
    case 2:
	// we have received everything for del profile already
	break;
	
	// Parameter update (4) or view (5) request
    case 4:
	
	if( _sendersSocket->get( update_len ) == false ) return FAIL;
	
	updatestring = new char[update_len+1];
	
	/* Get the client keyword string */
	memset(buf,0,BLOCK_SIZE);
	if( _sendersSocket->get_bytes( buf, update_len )!= update_len) return FAIL;
	memcpy(updatestring, buf, update_len);
	updatestring[update_len]='\0';
	
	log << "  Update/param view: \t"    << ((char*)updatestring)      << endl;
	
	break;
	
    default:
	break;
    }
    log << "---------------------------------------------------"  
	<< endl << endl;
    
    
    
    //------------------------------------------------------------------
    //  Process the new message according to the message type code
    //------------------------------------------------------------------
    
    //..............................
    //   A D D  
    //..............................
    if ( messageType == 1 ){
	
	int status=0;
	char cgiTranslationFile[100];
	sprintf (cgiTranslationFile, "%s%s", CGI_DIR, "translation.txt\0");
	
	
	// initialize a baddTanslator object (make a copy of the query string)
	baddTranslator BT( ap.id, ap.clientID, querystring );
	
	// translate the query and store the number of base tables
	int numTmp;
	char * tmp = BT.translateQuery( status , numTmp );
	
	//INFO("Trans:  "<<tmp<<endl);;
	
	
	// file to store translation so web client can display it
	ofstream trans(cgiTranslationFile);
	if(!trans){
	    cout <<"unable to open file "
		 << cgiTranslationFile << endl;
	    sendError(_sendersSocket, 1);
	    delete[] querystring;
	    freeProfile(ap);
	    return FAIL;
	}
	trans << tmp;
	trans.close();
	
	// set fields of the profile structure
	ap.numBaseTables = numTmp;
	ap.sqlString = strdup(querystring); 
	ap.dbname = strdup(IDS_DB_NAME);
	ap.deleted = false;                       // this profile is not deleted yet
	
	// init the time this profile was last run to be OLD (JAN 1, 1970)
	// so that the first time a new (or updated) profile is run it sends all relevant tuples
	ap.ts_last.tv_sec=0;  
	ap.ts_last.tv_nsec=0;	
	
	delete[] querystring;
	
	// Add or update this profile accordingly
	if(messageType == 1){ // Add
	    if( (st = addProfile( ap , tmp )) != OK){
		sendError( _sendersSocket, st);
		freeProfile(ap);
		return FAIL;
	    }
	}
    }      
    
    //..............................
    //   D E L E T E
    //..............................
    else if (messageType == 2){
	if((st = deleteProfile(ap.id, ap.clientID)) != OK ){
	    sendError( _sendersSocket, st);
	    return FAIL;
	}
    }
    
    //..............................
    //   P A R A M   U P D A T E 
    //..............................
    else if ( messageType == 4 ){
	if( (st = updateClientParameters( ap.id, updatestring )) != OK ){
	    sendError( _sendersSocket, st);
	    delete[] updatestring;
	    return FAIL;
	}
	delete[] updatestring;
    }
    
    //..............................
    //   P A R A M   V I E W
    //..............................
    else if ( messageType == 5 ){
	st = showClientParameters(ap.id, ap.clientID, updatestring );	
	if( st != 0 ){
	    _sendersSocket->put(st);  // pass the message on to message sender
	    _sendersSocket->close();
	    freeProfile(ap);
	    return FAIL;
	}
	delete[] updatestring;
    }
    
    _sendersSocket->put(0);
    _sendersSocket->close();
    
    freeProfile(ap);
    return OK;
}


//------------------------------------------------------------------------------------------
//  addProfile()  -  This method takes a generated profile (after translation)
//                   and adds it to the client profiles table.  If the profile  
//                   exists, an error is returned to caller
//
//  ARGS:     ap - the profile to add
//            translation - the translation
//
//  RETURNS:  error code { OK, DB_OPEN_ERROR, PROFILE_EXISTS, CURSOR_ERROR }
//------------------------------------------------------------------------------------------
int 
pullReceiver::addProfile(AProfile ap, char* translation)
{
    
    int         st, numTuples=0;
    ostrstream  os, os2, os3;
    
    os << "select * from MY_PROFILES P where P.PID="<<ap.id
       << " and p.cid="<<ap.clientID<<ends;
    
    //INFO("Adding a profile");
    
    if( ( st = _query->execute(os.str(), numTuples))!=OK){
	ERROR("Cursor Error: "<< st);
	os.freeze(0);
	return CURSOR_ERROR;
    }
    if (numTuples != 0){
	WARN("Profile id "<<ap.id<<" already exists for client "<<ap.clientID ); 
	os.freeze(0);
	return (PROFILE_EXISTS);   // code 11
    }
    
    os.freeze(0); // unfreeze the stream

    cout << "\n\nRECEIVER THREAD ADD: (sql) "<<ap.sqlString<<endl;
    cout << "RECEIVER THREAD ADD: (pid) "<<ap.id<<endl;
    cout << "RECEIVER THREAD ADD: (cid) "<<ap.clientID<<endl;
    cout << "RECEIVER THREAD ADD: (adr) "<<ap.clientAddr<<endl;
    cout << "RECEIVER THREAD ADD: (dnm) "<<ap.dbname<<endl;
    cout << "RECEIVER THREAD ADD: (per) "<<ap.period<<endl;
    cout << "RECEIVER THREAD ADD: (tsl) "<<ap.ts_last.tv_sec<<endl;
    cout << "RECEIVER THREAD ADD: (por) "<<ap.port<<endl;
    cout << "RECEIVER THREAD ADD: (nbt) "<<ap.numBaseTables<<endl<<endl;
    
    ////////////////////////////////////////////////////////////////////////////////////////
    //
    //  Construct an insert stmt for the client side profile table.
    //  The schema of this table is as follows:
    //   
    //   field      name                type        desc
    //   -----      -------------       --------    ------------------------------------
    //     0        pid                 Integer     This profile ID
    //     1        cid                 Integer     Our ID
    //     2        cliaddr             Integer     Our Address
    //     3        profile             String      The profile SQL string
    //     4        translation         String      The translated profile SQL string
    //     5        db_name             String      
    //     6        period              Integer     Period between profile executions 
    //     7        init_time           Integer     The initial Timestamp (normally 0)
    //     8        my_port             Integer     Our incoming port (for data from IDS)
    //     9        numTempTables       Integer     The number of tables in the from clause
    //     10       prof_state          Integer     Has this profile been registered at IDS
    //
    //
    /////////////////////////////////////////////////////////////////////////////////////////
    os2 << "insert into MY_PROFILES values ("<<ap.id<<", "<<ap.clientID<<", '"
	<< ap.clientAddr << "', '" << ap.sqlString <<"', '"<< translation <<"', '"
	<< IDS_DB_NAME <<"', "<< ap.period << ", 0 , " << ap.port<<", "<< ap.numBaseTables
	<< ", 1)" << ends;
    
    
    if( (st = _query->execute(os2.str(), numTuples)) != OK){
	ERROR("CURSOR_ERROR: " << os2.str());
	os2.freeze(0);
	return CURSOR_ERROR;
    }
    
    if (numTuples != 1){
	WARN( "Insertion of profile id "<<ap.id<<" for client "<<ap.clientID<< " failed!"); 
	WARN( "Query: "<<os2.str() );
	os2.freeze(0);
	return (ADD_FAILED);
    }

    os2.freeze(2); // unfreeze the stream    
    return OK;
}


//------------------------------------------------------------------------------------------
//  deleteProfile()  -  Change the stat of a profile to deleted so the updater thread will 
//                      delete it at the IDS site
//
//  ARGS:     pid - profile id
//            cid - client id
//
//  RETURNS:  error code { OK, DB_OPEN_ERROR, FAIL, NO_SUCH_PROFILE }
//------------------------------------------------------------------------------------------
int 
pullReceiver::deleteProfile( int pid, int cid )
{
       int err, i, numTuples = 0;
       ostrstream os;
       
       //------------------------------------------------------------
       // Update this profile's state to be "deleted", IDS updater 
       // thread will handle the rest
       //------------------------------------------------------------
       os << "update MY_PROFILES set PROF_STATE = 2 where PID="<<pid
	  << " and cid="<< cid <<ends;
       
       if( _query->execute( os.str(), numTuples ) != OK){
	    WARN("Query execution failed: " << os.str() );
	    os.freeze(0);
	    return (FAIL);
       } 
       os.freeze(0);

       if (numTuples == 0){
	    return (NO_SUCH_PROFILE);
       }       

       return OK;
}


//------------------------------------------------------------------------------------------
//  updateClientParameters()  -  Execute a parameter add/update stmt locally and add the
//                               stmt to the ids_updates table for the updater thread to 
//                               execute the stmt at the ids_site
//  ARGS:     pid - the profile id
//            sqlstring - the statement to execute
//
//  RETURNS:  error code { OK, FAIL, DB_OPEN_ERROR, FAIL } 
//------------------------------------------------------------------------------------------
int  
pullReceiver::updateClientParameters( int pid, char* sqlstring )
{
       int ret, i, numTuples = 0;
       ostrstream os;
       ostrstream sqlstmt;
       char * ptr;

       //-----------------------------------------------------
       //  Execute the sql statement
       //-----------------------------------------------------
       ret = _query->execute(sqlstring, numTuples); 
       if(ret != OK ){
	    WARN("Execution of query failed: "<< sqlstring );
	    return FAIL;
       }
       if (numTuples == 0){
	   WARN("Param update stm had no effect, (no tuples affected) "<<sqlstring);
	   return FAIL;
       }

       // Escape the string delimiters
       for(ptr = sqlstring; *ptr; ptr++){
	   if(*ptr == '\''){
	       //INFO("matched a ' !");
	       sqlstmt << "\\" << *ptr;
	   }else{
	       sqlstmt << *ptr;
	   }
       }       
       //INFO("Here is the sqlstmt: "<<sqlstmt.str());

       //--------------------------------------------------
       // Insert the string to IDS update table
       //--------------------------------------------------
       os << "INSERT INTO IDS_UPDATES VALUES ("<<pid<<", '"<<sqlstmt.str()<< "'," <<1<<")"<<ends;
       
       if( (ret = _query->execute( os.str(), numTuples )) != OK){
	    WARN("Error inserting to ids_updates: "<< sqlstring);
	    sqlstmt.freeze(0);
	    os.freeze(0);
	    return FAIL;
       }

       sqlstmt.freeze(0);
       os.freeze(0);
       return OK;
}


//------------------------------------------------------------------------------------------
//  showClientParameters()  -  Show the parameters we have registered at the local site
//
//  ARGS:     sock - the connection to send error to 
//            errorcode - the error code to send     
//  RETURNS:  error code { OK, DB_OPEN_ERROR, FAIL }
//------------------------------------------------------------------------------------------
int  
pullReceiver::showClientParameters( int cid, int pid, char * sqlstring )
{
       int st, i;
       char filename[100];
       
       sprintf(filename, "%s/%d.%d.param", CGI_DIR, pid, cid);

       //--------------------------------------------------
       //  Dump query result to file, close db
       //--------------------------------------------------
       if( _query->html_table_dump(sqlstring, filename) != OK) { 
	    WARN("HTML table dump failed"); 
	    return FAIL;
       }
       return OK;
}


//------------------------------------------------------------------------------------------
// sendError()  -  Respond to an error condition.  Send the error code to the
//                 client connection, close the connection, display a message
//
//  ARGS:     sock - the connection to send error to 
//            errorcode - the error code to send     
//  RETURNS:  none
//------------------------------------------------------------------------------------------
void 
pullReceiver::sendError(TCPSock* sock, int errorcode)
{
       BEM.display(errorcode);
       sock->put(errorcode);  // pass the message on to message sender
       sock->close();  
}












