#include <readWriteLock.h>

//----------------------------------------------------------------------------------------------------
//  constructor()  -  initialize the reader/writer lock
//
//  ARGS:     numreaders - number of readers (default is 0)
//            numwriters - number of writers (default is 0)
//
//  RETURNS:  none
//----------------------------------------------------------------------------------------------------
readWriteLock::readWriteLock( int numreaders, int numwriters )
{
     readersReading = numreaders;
     writersWriting = numwriters;
     pthread_mutex_init(&rw_mutex, NULL);
     pthread_cond_init(&lock_free, NULL);
}
 

//----------------------------------------------------------------------------------------------------
//  destructor()  -  free the mutex and convar
//
//  ARGS:     none
//  RETURNS:  none
//----------------------------------------------------------------------------------------------------
readWriteLock::~readWriteLock( )
{
    pthread_mutex_destroy( &rw_mutex );
    pthread_cond_destroy( &lock_free );
}

//----------------------------------------------------------------------------------------------------
//  readerLock()  -  Set a reader lock
//
//  ARGS:     none
//  RETURNS:  none
//----------------------------------------------------------------------------------------------------
void 
readWriteLock::readerLock()
{
     pthread_mutex_lock(&rw_mutex);
     while(writersWriting>0){
#ifdef RW_DEBUG
          cout << "\t**********************************\n";
	  cout << "\t*  READER WAITING ON WRITER  R: "<<readersReading<<" W: "<<writersWriting<<endl;
	  cout << "\t**********************************\n";
#endif
	  pthread_cond_wait(&lock_free, &rw_mutex);
     }
     readersReading++;
#ifdef RW_DEBUG
     cout << "\t**********************************\n";
     cout << "\t*  READER OBTAINED LOCK  R: "<<readersReading<<" W: "<<writersWriting<<endl;
     cout << "\t**********************************\n";
#endif
     pthread_mutex_unlock(&rw_mutex);
}


//----------------------------------------------------------------------------------------------------
//  writerLock()  -  Set an exclusive writer lock
//
//  ARGS:     none
//  RETURNS:  none
//----------------------------------------------------------------------------------------------------
void 
readWriteLock::writerLock()
{
     pthread_mutex_lock(&rw_mutex);
     while( (writersWriting>0) || (readersReading>0) ){
#ifdef RW_DEBUG
          cout << "\n\t**********************************\n";
	  cout << "\t*  WRITER WAITING....  R: "<<readersReading<<" W: "<<writersWriting<<endl;
	  cout << "\t**********************************\n";
	  pthread_cond_wait(&lock_free, &rw_mutex);
#endif
     }
     
     writersWriting++;
#ifdef RW_DEBUG
     cout << "\n\t**********************************\n";
     cout << "\t*  WRITER GOT LOCK  R: "<<readersReading<<" W: "<<writersWriting<<endl;
     cout << "\t**********************************\n";
#endif
     pthread_mutex_unlock(&rw_mutex);
}


//----------------------------------------------------------------------------------------------------
//  readerUnlock()  -  Decrement the number of reader locks set
//
//  ARGS:     none
//  RETURNS:  errorcode - negatinve if no readers
//----------------------------------------------------------------------------------------------------
int 
readWriteLock::readerUnlock()
{
     pthread_mutex_lock(&rw_mutex);

     /////////////////////////////////////////////
     //
     //  If there are no readers, return an error
     //
     /////////////////////////////////////////////
     if(readersReading == 0){
       pthread_mutex_unlock(&rw_mutex);
       return -1;
     }
     
     /////////////////////////////////////////////
     //
     //  Decrement the readers variable and 
     //  signal condition if it goes to 0
     //
     /////////////////////////////////////////////
     else{       
       readersReading--;
       if(readersReading==0){
#ifdef RW_DEBUG
	 cout << "\n\t**********************************\n";
	 cout << "\t*  ALL READERS OUT  R: "<<readersReading<<" W: "<<writersWriting<<endl;
	 cout << "\t**********************************\n";
#endif	   
	 pthread_cond_signal(&lock_free);
       }
#ifdef RW_DEBUG
       cout << "\n\t**********************************\n";
       cout << "\t*  READER RELEASED LOCK  R: "<<readersReading<<" W: "<<writersWriting<<endl;
       cout << "\t**********************************\n";
#endif
       pthread_mutex_unlock(&rw_mutex);
       return 0;
     }
}


//----------------------------------------------------------------------------------------------------
//  writerUnlock()  -  Release the exclusive writers lock
//
//  ARGS:     none
//  RETURNS:  error code - negative if no writers writing
//----------------------------------------------------------------------------------------------------
int 
readWriteLock::writerUnlock()
{
     pthread_mutex_lock(&rw_mutex);

     /////////////////////////////////////////////
     //  If there is no writer, return an error
     /////////////////////////////////////////////
     if(writersWriting==0){
       pthread_mutex_unlock(&rw_mutex);
       return -1;
     }
     
     /////////////////////////////////////////////
     //  Set the writers variable to 0 and 
     //  broadcast to those waiting on condition var
     /////////////////////////////////////////////
     else{       
          writersWriting=0;
	  pthread_cond_broadcast(&lock_free);
#ifdef RW_DEBUG 
	  cout << "\n\t**********************************\n";
	  cout << "\t*  WRITER IS OUT  R: "<<readersReading<<" W: "<<writersWriting<<endl;
	  cout << "\t**********************************\n";
#endif
	  pthread_mutex_unlock(&rw_mutex);
	  return 0;
     }
}









