//////////////////////////////////////////////////////////////////////////////////////////
//
//   sql_stuff.cpp
//
//   Brucej Jackson,     modified from: Sowmya Subramanian
//   Prof. DeWitt
//
//   This file contains all the methods for the SQLQuery class. These methods are
//   basic cursor operations that are used as interfaces to the paradise server.
//   The different methods include, query execute, schema dump, update query,
//   and query execute and dump.
//
//////////////////////////////////////////////////////////////////////////////////////////

#include <sql_stuff.h>


//----------------------------------------------------------------------------------------
//   constructor() -  Constructs a query object attached to no specific DB
//
//   ARGS:      none
//   RETURNS:   none
//----------------------------------------------------------------------------------------
SQLQuery::SQLQuery()
{
    _hostname = getenv("PDHOST");
    if(!_hostname){
	_hostname = "localhost";
    }
    _port = PARADISE_PORT;
    _dbname = NULL;
    _connection = new ParaConnect(_hostname,_port);
    if(_hostname){
	delete(_hostname);
    }
    if(! _connection ){
	ERROR("Failed to create new connection object");
	return;	
    }
    
    _cursor = new ParaCursor(*_connection, ParaErrorStatus::eDetailed);
    if(! _cursor ){
	ERROR("Failed to create new cursor object");
	delete _connection;
	return;
    }
    int err = _cursor->authenticate(TOR_USER, TOR_PASSWORD);	
    if(err != clNOERROR){
	ERROR( "Authentication failed, aborting" );
	//exit(1);
    }
}

//----------------------------------------------------------------------------------------
//   constructor() - Constructs a query object associated with the db given as and arg
//
//   ARGS:      dbname - the db to associate this query object with
//              port - the port to connect to
//   RETURNS:   none 
//----------------------------------------------------------------------------------------
SQLQuery::SQLQuery(const char *dbname, int port)
{
    _connection = NULL;
    _cursor = NULL;
    
    _hostname = getenv("PDHOST");
    if(!_hostname)
	_hostname = "localhost";
    _port = port;
    _dbname = strdup(dbname);
    
    _connection = new ParaConnect(_hostname,_port);
    if(! _connection){
	ERROR("Failed to create new connection object");	
	return;
    }
    _cursor = new ParaCursor(*_connection);
    if(! _cursor){
	ERROR("Failed to create new cursor object");
	return;
    } 
    int err = _cursor->authenticate(TOR_USER, TOR_PASSWORD);
    if(err != clNOERROR){
	ERROR("Authentication failed, aborting");
	//exit(1);
    }
}


//----------------------------------------------------------------------------------------
//   destructor() - frees memory associated with this query object
//
//   ARGS:      none
//   RETURNS:   none
//----------------------------------------------------------------------------------------
SQLQuery::~SQLQuery()
{
    _cursor->close();
    _connection->close();	
    if( _cursor ) delete _cursor;     
    if( _connection ) delete _connection;
    if( _dbname ) free( _dbname );
}


//----------------------------------------------------------------------------------------
//   open_db()
//
//   ARGS:      none
//   RETURNS:   error code: { OK, DB_OPEN_ERROR } 
//----------------------------------------------------------------------------------------
int 
SQLQuery::open_db()
{
    //int st;
    char queryStr[DEFAULT_LEN];
    
    sprintf(queryStr, "OPEN DB %s", _dbname);
    CURSOR_CALL2( _cursor->declare(queryStr), DB_OPEN_ERROR, false );
    CURSOR_CALL2( _cursor->exec(), DB_OPEN_ERROR, false );
    CURSOR_CALL2( _cursor->close(), DB_OPEN_ERROR, false );
    
    return (OK);
}


//----------------------------------------------------------------------------------------
//   close_db()
//
//   ARGS:      none
//   RETURNS:   error code: { OK, DB_CLOSE_ERROR } 
//----------------------------------------------------------------------------------------
int 
SQLQuery::close_db()
{
    //int st;
    char queryStr[DEFAULT_LEN];
    sprintf(queryStr, "CLOSE DB %s", _dbname);
    
    CURSOR_CALL2( _cursor->declare(queryStr), DB_CLOSE_ERROR, false );
    CURSOR_CALL2( _cursor->exec(), DB_CLOSE_ERROR, false );
    CURSOR_CALL2( _cursor->close(), DB_CLOSE_ERROR, false );
    
    return OK;
}


//----------------------------------------------------------------------------------------
//   table_exists() - Checks if a table exists by querying the sys_relations table.
//                    Assumes the db is open when called and closed by the caller
//
//   ARGS:      table_name - the name of the table to check
//   RETURNS:   error code: { EXISTS, NOT_EXISTS, FAIL } 
//----------------------------------------------------------------------------------------
int 
SQLQuery::table_exists(char* table_name)
{
    char queryStr[DEFAULT_LEN];
    int st, numTuples=0;
    
    sprintf(queryStr, "SELECT * FROM SYS_RELATIONS S WHERE S.NAME='%s'", table_name);
    
    if( (st = _cursor->declare(queryStr)) != clNOERROR) return FAIL;
    if( (st = _cursor->exec()) != clNOERROR) return FAIL;
    if( (st = _cursor->numTupleAffected(numTuples)) != clNOERROR) return FAIL;
    
    if(numTuples>0)
	return (EXISTS);
    return (NOT_EXISTS);
}


//----------------------------------------------------------------------------------------
//   execute_and_dump() -  This method executes a query and dumps the schema info to a 
//                         schema file, and the data into a data file. 
//                    
//   ARGS:      query - the query to execute
//              prof_dir - the working directory for this profile
//              numTables - the number of base tables in this profile
//              curNum - the table number being dumped in this call
//              cid - the client is for the profile
//              pid - the profile id of the profile
//              
//   RETURNS:   error code: { OK, NO_RESULT, -1 } 
//----------------------------------------------------------------------------------------
int 
SQLQuery::execute_and_dump( char *query, char *prof_dir, int numTables, int curNum, int cid, int pid)
{
    char                      queryStr[DEFAULT_LEN];
    char                      schemaFile[100];
    char                      tableName[100];
    char                      filename[100];
    //ostrstream                temp1, temp2, temp3;
    ClientBaseAdt**           result;
    //BaseAdtInfo::BaseAttrType attrType;
    char                      bFilename[100];
    int                       ret, numTuples, st, i;
    schemaState               state;
    
    // determine what SQL statements we will need 
    // to generate in the schema file during this call
    if( curNum==1 && numTables==1 )
	state=GEN_BOTH;
    else if( curNum==1 && numTables>1 )
	state=GEN_OPEN;
    else if( curNum < numTables )
	state=GEN_NONE; 
    else if( curNum == numTables )
	state=GEN_CLOSE; 
    
    
    // generate the schema file name, 1 schema file
    // per profile, but may be multiple profile files (one 
    // for each base table in the profile query
    sprintf(schemaFile,"%s/%d.%d.schema", prof_dir, cid, pid);
    sprintf(filename, "%s/%d.%d.out.%d", 
	    prof_dir, cid ,pid, curNum );
    
    // Extract the base table name from the query
    char* ptr = query + 14;  //bypass the 14 char 'select * from ' part of this query
    i=0;
    while(*ptr != ','){
	if(*ptr != ' '){
	    tableName[i]=*(ptr);
	    i++;
	}
	ptr++;
    }
    tableName[i]='\0';
    
    
    // bind and execute the query
    sprintf(queryStr, query);
    CURSOR_CALL2( _cursor->declare( queryStr ), 0 , false );
    CURSOR_OPEN_CALL( _cursor->open() );
    
    
    /*
     * for the execute and dump function, we are expecting there to be a result
     * unless all tuples are old.  In this case we do not want to create an
     * empty file and disseminate useless information to the client.  Therefore,
     * we check each time for a result before dumping, disseminating.
     */
    
    CURSOR_CALL2( _cursor->numTupleAffected( numTuples ), 0, false );
    if( numTuples == 0 ) {
	//INFO( "NO RESULT TUPLES, no output to dissemination file");
	CURSOR_CALL( _cursor->close() );
	return NO_RESULT;  // signal caller that there was no result
    }
    
    // otherwise execute and dump
    else{
	
	CURSOR_CALL( _cursor->fetch() );
	
	// get the number of attributes in the result schema
	int attrCount;
	CURSOR_CALL( _cursor->numAttrs(attrCount) );
	
	// dump schema information to schema file, then dump tuples to data file
	if( (ret = dump_schema(schemaFile, tableName, filename, attrCount, 
			       numTables, state)) != OK ){
	    ERROR("Failed while dumping schema: "<<ret);
	}
	
	// open the DATA file to dump tuples to
	ofstream outf(filename, ios::out);
	if(! outf ) {
	    ERROR( "File open error: " << filename << endl);
	    CURSOR_CALL( _cursor->close());
	    return -1;
	}
	
	
	int bfCnt=0;
	int tplCnt=1;	  
	do{
	    
	    // get the next tuple and dump the data 
	    // for each field of this tuple to the data file
	    // if a field in this tuple is a file (image, text doc etc)
	    // then it's data is dumped to a seperate binary file
	    // and the file name is recorded in the data file
	    CURSOR_CALL( _cursor->getTuple(result) );
	    
	    
	    // Dump each field, if the field type that will be written
	    // is a file, the memToAscii function will use the bFilename to 
	    // create the binary file (attrCount-2 so we do not send the joining garbage TS)
	    for (i=0; i<attrCount-2; i++){
		sprintf( bFilename, "%s.bf%d",filename, bfCnt);
		bfCnt++;
		if ( result[i]->memToAscii(outf, bFilename) ){
		    CURSOR_CALL( _cursor->close() );
		    ERROR("Failed while calling memToAscii function during dump");
		    outf.close();
		    return (-1);
		} 
		outf << endl;
	    }
	    
	    // Add a blank line at the end of this tuple, 
	    // and fetch the next tuple
	    outf << endl;
	    ret = _cursor->fetch();
	    tplCnt++;
	} while ( !ret );	  
	
	// make sure we ended on a valid error condition
	if ( ret != clEND_TUPLES ) {
	    ERROR( cem[ret].message<< " ("<<ret<<")");
	    CURSOR_CALL( _cursor->close() );
	    return (-1);
	    outf.close();
	}
	
	// query executed and dumped successflly
	outf.close();
	CURSOR_CALL( _cursor->close() );
	return OK;
    }	
}

//----------------------------------------------------------------------------------------
//  execute() - This method uses the cursor to perform execute only queries.  If a 
//              result relation is expected, then use execute_and_dump method above 
//              This method is used for things like 'open db', 'create table'
//
//  ARGS:     query - the query to execute
//            numTuples - by reference argument to return the number of tuples affected
//                        by the execution of this query
//
//  RETURN:   error code:  { OK, CURSOR_ERROR }
//----------------------------------------------------------------------------------------
int 
SQLQuery::execute(char * query, int& numTuple)
{  
    numTuple=0;   
    char  queryStr[DEFAULT_LEN];
    int   ret, err;
    
    sprintf( queryStr,"%s", query );
    
    CURSOR_CALL2( _cursor->declare( queryStr ) , CURSOR_ERROR , false );
    CURSOR_CALL2( _cursor->exec(), CURSOR_ERROR, false );
    CURSOR_CALL2( _cursor->numTupleAffected(numTuple), CURSOR_ERROR, false );
    CURSOR_CALL2( _cursor->close(), CURSOR_ERROR, false );
    
    return OK;
}



//----------------------------------------------------------------------------------------
//  dumpSchema() - This method generates the SQL statements used to load the db on the 
//                 client (smartPull) side.  These statements are dumped to the schema
//                 file that is send with the data to the smartPull client
//
//  ARGS:    schemaFile - 
//           tableName - 
//           dataFile - 
//           attrCount - 
//           numTables - 
//           schemaState - 
//
//  RETURN:  error code { OK }
//----------------------------------------------------------------------------------------
int 
SQLQuery::dump_schema( char* schemaFile, char* tableName, char* dataFile, int attrCount, 
		       int numTables, schemaState state)
{
    const char *    attrName;
    const char *    attrAdtName;
    ostrstream      fs;          // name of this file
    ostrstream      dirname;     // directory name
    ostrstream      name;        //  	
    //ostrstream      attr,junk,tbname;
    //ostrstream      os;
    char *          tmpPtr;
    BaseAdtInfo::BaseAttrType attrType;
    
    
    
    // open schemaFile in append mode.  This file will be deleted 
    // by a profile after the profile has finished (and after this 
    // schema file has been sent to the client
    ofstream outf(schemaFile, ios::app);	
    
    //INFO( "Dumping to schemafile: "<<schemaFile);
    
    
    // print numTables for this prof and
    // generate an open db statement on the first call
    if( (state == GEN_OPEN) || (state == GEN_BOTH) ){
	outf << numTables <<endl;	
	outf << "OPEN DB CL" << _dbname << endl;
    }
    
    //------------------------------------------------------------------------
    // Generate the table name, create table, and load table statements
    // NOTE: I am using attrCount-2 here because there is no need to send the 
    // fields that join with the base table (e.g. The ID field, which is 
    // the same as ID, and the minOfOld field used to determine what to 
    // disseminate) 
    //------------------------------------------------------------------------
    for(int i=0; i<attrCount-2; i++)
    {
	
	// get the ADT name and Attr Name for this attr
	// after these calls, memory has been allocated for 
	// these two char arrays
	CURSOR_CALL(_cursor->attrType(i, attrType));
	attrAdtName = ClientAdtInfo::getAdtName(attrType);
	CURSOR_CALL(_cursor->attrName(i, attrName));
	
	// the attrName is of the form: <attrname>$<tabname>
	// Extract name of attribute and table name.			
	char * tmpAttrName = strdup(attrName);
	ostrstream attr,junk,tbname;
	split_string(tmpAttrName,attr, tbname, '$'); 
	attr.seekp(0);
	split_string(attr.str(),attr,junk,'^');
	attr.seekp(0);
	split_string(attr.str(),attr,junk,'.');

	free(tmpAttrName);
	
	// Construct the create table <tname>(attr,...) stmt 
	ostrstream os;
	if(i==0) {
	    // on the first pass, generate the create table stmt.
	    os <<tableName <<"\nCREATE TABLE " << tableName 
	       << " ( " 
	       << attr.str() << " ";
	    os << attrAdtName;
	    os << ends;
	    outf << os.str();
	    
	    os.freeze(0);
	    attr.freeze(0);
	    junk.freeze(0);
	    tbname.freeze(0);	
	    continue;
	}
	
	os << ", " << attr.str() << " "
	   << attrAdtName << ends;
	outf << os.str();
	
	// cleanup
	os.freeze(0);
	attr.freeze(0);
	junk.freeze(0);
	tbname.freeze(0);
    }
    
    outf << ")" << endl;
    outf << "LOAD TABLE " << tableName << " FROM \"" << dataFile 
	 << "\" " << endl;
    
    
    // generate a close db statment if it is needed
    if(state==GEN_CLOSE || state==GEN_BOTH){
	outf << "CLOSE DB CL" << _dbname << endl;
    }
    
    outf.close();	
    
    fs.freeze(0);          // name of this file
    dirname.freeze(0);     // directory name
    name.freeze(0);        //  	
    return OK;
}



//----------------------------------------------------------------------------------------
//  run_update() - Update a a profiles TS field in the DB (time the profile was last run)
//
//  ARGS:     pid - profile id
//            cid - client id
//            runtime - time profile was executed last
//
//  RETURN:   error code:  { OK }
//----------------------------------------------------------------------------------------
int 
SQLQuery::run_update(int pid, int cid, int runtime)
{
    char   queryStr[DEFAULT_LEN];
    int    ret;
    int    numTuples = 0;
    int    numTries  = 0;
    
    sprintf(queryStr, "UPDATE PROFILES SET TS = %d WHERE CID = %d AND ID = %d", runtime, cid, pid);
    
    //INFO("executing update query: "<<queryStr);
    
    do{
	ret = execute(queryStr, numTuples);
	if(ret != OK){
	    sleep(2);
	    if(numTries++ < 10)
		return UPDATE_FAILED;
	    WARN("Failed to exec update, code is: "<<ret<<", attemptNum: "<<numTries); 
	}
    }while (ret != OK);
    
    return OK;
}


int 
SQLQuery::html_table_dump( char* queryStr, char* filename )
{
    ClientBaseAdt** result;
    BaseAdtInfo::BaseAttrType attrType;
    
    int      st, i;
    int      ret;
    int      numTuples=0;
    char *   attrName;

    ofstream outf(filename, ios::out);
    if(!outf){
	cerr <<"failed to open file"<<endl;
	return FAIL;
    }
    
    CURSOR_CALL( _cursor->declare(queryStr));
    CURSOR_OPEN_CALL( _cursor->open());
    ret = _cursor->fetch();
    CURSOR_CALL( _cursor->numTupleAffected(numTuples));
    
    if(numTuples == 0){
	outf << "No Result!, make sure you have added the parameters you are trying to modify"<<endl;
	return FAIL;
    }
    
    // get the number of attributes in the result schema
    int attrCount;
    CURSOR_CALL(_cursor->numAttrs(attrCount));
    
    outf << "<TABLE BORDER=1 CELLSPACING=2 CELLPADDING=2 BGCOLOR=\"#CCCCCC\" WIDTH=75%>"
	 <<endl;
    outf << "<CAPTION><FONT SIZE=2><B>Paradise Query Result</B></FONT></CAPTION>"<<endl;
    outf << "<TR>";
    for (i=0; i<attrCount; i++){
	
	CURSOR_CALL( _cursor->attrName(i, attrName));
	
	// the attrName is of the form: <attrname>$<tabname>
	// Extract name of attribute and table name.			
	ostrstream attr,junk,tbname;
	split_string((char *)attrName,attr, tbname, '$'); 
	attr.seekp(0);
	split_string(attr.str(),attr,junk,'^');
	attr.seekp(0);
	split_string(attr.str(),attr,junk,'.');
	outf << "<TH>" <<attr.str()<<"</TH>";
	attr.freeze(0);
	junk.freeze(0);
	tbname.freeze(0);
    }
    
    outf << "</TR>" <<endl;
    
    do{
	CURSOR_CALL(_cursor->getTuple(result));
	outf << "<TR BGCOLOR = \"white\">";
	// Dump each field , tab delimited, (no bin files here)
	for (i=0; i<attrCount; i++){
	    outf << "<TD>";
	    if ( result[i]->memToAscii(outf) ){
		_cursor->close();
		cerr << "**ERROR: calling memToAscii function during dump failed!.\n";
		outf.close();
		return FAIL;
	    } 
	    outf << "</TD>";
	}
	outf << "</TR>"<<endl;
	
	ret = _cursor->fetch();
    } while ( !ret );		
    
    outf << "</TABLE><br>";
    return OK;
}


/////////////////////////////////////////////////////////////////////////////////
//
//  Was used when profiles were registered at the main IDS rather than the client
//
//  This method exports the entire badd schema including all
//  Parameter, Data, and Profile table schemas.  It writes the
//  schema information in HTML tbale format to a file that is 
//  read by a the Web CGI client.  This function is called each 
//  time a client wishes to view the schema.  This is not efficient, but
//  may be changed to be run on a periodic basis, updating the schema 
//  information.  This method, as with other (SQLQuery methods) assumes there 
//  is a db open when called
//
////////////////////////////////////////////////////////////////////////////////
// int SQLQuery::export_badd_schema()
// {
//        int st, i, j, k, numTuples, numTables, attrCount, ret;
//        char schemaFile[100];
//        char qryString[256];
//        bool append = true;
//        ClientBaseAdt** result;
//        BaseAdtInfo::BaseAttrType attrType;
//        ClientBaseAdt *obj;
//        char* attrName;
       
//        struct tableInfoTuple{
// 	 char* tname;
// 	 char* ttype;
// 	 char* tdesc;
//        };
       
//        sprintf(schemaFile, "%s/%s", CGI_DIR, "badd.schema");
       
//        /////////////////////////////////////////////////////////////////////////
//        //
//        //  Export BADD Table Schemas for all Parameter, profile and data tables
//        //
//        /////////////////////////////////////////////////////////////////////////

//        sprintf(qryString, "%s", "select * from BADD_Tables where table_type=\"profile\" or table_type=\"parameter\" or table_type=\"data\"\0");
       
//        CURSOR_CALL( _cursor->declare(qryString) );
//        CURSOR_CALL( _cursor->open() );
//        CURSOR_CALL( _cursor->numTupleAffected(numTables) );

//        tableInfoTuple tables[numTables];
//        int tblcount=0;

//        if(numTables==0){
// 	 return FAIL;
//        }

//        CURSOR_CALL(_cursor->numAttrs(attrCount));
	   

//        // Store the BADD Table data in a in an array
//        // of table description structs for now
//        while( !(ret = _cursor->fetch()) ){
// 	 CURSOR_CALL(_cursor->getTuple(result));
	 
// 	 // Save a local copy of each tuple
// 	 for (i=0; i<attrCount; i++){
	   	   
// 	   CURSOR_CALL(_cursor->getObj(obj,attrType,i));
// 	   switch (i) {
	     
// 	   case 0:                       // tablename
// 	     tables[tblcount].tname = 
// 	       strdup((const char*)((ClientStringAdt*)obj)->getStrVal());
// 	     break;
// 	   case 1:                       // table type
// 	     tables[tblcount].ttype = 
// 	       strdup((const char*)((ClientStringAdt*)obj)->getStrVal());
// 	     break;
// 	   case 2:                       // table description
// 	     tables[tblcount].tdesc = 
// 	       strdup((const char*)((ClientStringAdt*)obj)->getStrVal());
// 	     break;
// 	   default:
// 	     break;
// 	   }
// 	 }

// 	 tblcount++;
//        } 


//        // Open the schema file
//        ofstream outf;
//        outf.open(schemaFile, ios::out);
//        if(!outf){
// 	 // free up mem
// 	 for(i=0; i<numTables; i++){
// 	   if (tables[i].tname)
// 	     free( tables[i].tname);
// 	   if (tables[i].ttype)
// 	     free( tables[i].ttype);
// 	   if (tables[i].tdesc)
// 	     free( tables[i].tdesc);
// 	 }
// 	 return FAIL;
//        }       


//        //////////////////////////////////////////////////////////////
//        //
//        //  For each table, export it with caption in HTML table format
//        //
//        //////////////////////////////////////////////////////////////
//        for (i=0; i<numTables; i++){

// 	     // Get the system catalog info for this table
// 	     sprintf(qryString, "select r.name, a.name, t.name from sys_relations r, sys_types t, sys_attributes a where a.tupleTypeId=r.typeId and a.attrTypeId=t.typeId and r.name=\"%s\"\0", tables[i].tname);

// 	     //cout <<"Binding query: "<<qryString<<endl;
// 	     cout <<"EXPORT SCHEMA: "<<tables[i].tname<<endl;

// 	     CURSOR_CALL( _cursor->declare(qryString) );
// 	     CURSOR_CALL( _cursor->open() );
// 	     CURSOR_CALL( _cursor->numTupleAffected(numTuples) );
	     
// 	     // get the number of attributes in the result schema
// 	     CURSOR_CALL(_cursor->numAttrs(attrCount));
	     
// 	     // return false
// 	     if(numTuples==0){
// 	       // free up mem
// 	       for(i=0; i<numTables; i++){
// 		 if (tables[i].tname)
// 		   free( tables[i].tname);
// 		 if (tables[i].ttype)
// 		   free (tables[i].ttype);
// 		 if (tables[i].tdesc)
// 		   free (tables[i].tdesc);
// 	       }
// 	       return FAIL;
// 	     }	     
	     
// 	     // Print a header for this HTML table
// 	     outf << "<TABLE BORDER=1 CELLSPACING=2 CELLPADDING=2 WIDTH=75%>"<<endl;
// 	     outf << "<CAPTION><B>Relation Name: <FONT SIZE=3 COLOR=\"blue\">"<<  tables[i].tname 
// 		  << " </FONT></B><I>("<< tables[i].ttype 
// 		  <<" table)</I><br>"<<tables[i].tdesc <<"</CAPTION>"<<endl;
	     
// 	     outf << "<TR BGCOLOR=\"#CCCCCC\"><TH>Attribute Name</TH><TH>ADT Type</TH></TR>";
	     
	     
// 	     //Generate the data section of the table
// 	     ret = _cursor->fetch();
// 	     do{
// 	         CURSOR_CALL(_cursor->getTuple(result));
// 		 outf << "<TR BGCOLOR = \"white\">";
// 		 // Dump each field , tab delimited, (no bin files here)
// 		 for (k=1; k<attrCount; k++){
// 		   outf << "<TD>";
// 		   if ( result[k]->memToAscii(outf) ){
// 		     _cursor->close();
// 		     cerr << "**ERROR: calling memToAscii function during dump failed!.\n";
// 		     outf.close();

// 		     // free up mem
// 		     for(i=0; i<numTables; i++){
// 		       if (tables[i].tname)
// 			 free( tables[i].tname);
// 		       if (tables[i].ttype)
// 			 free( tables[i].ttype);
// 		       if (tables[i].tdesc)
// 			 free( tables[i].tdesc);
// 		     }
// 		     return FAIL;

// 		   } 
// 		   outf << "</TD>";
// 		 }
// 		 outf << "</TR>"<<endl;
// 		 ret = _cursor->fetch();
// 	     }while ( !ret );		
// 	     outf << "</TABLE><BR><BR>";
	     
//        }// End of table schema dump


//        // free up mem
//        for(i=0; i<numTables; i++){
// 	 if (tables[i].tname)
// 	   free( tables[i].tname);
// 	 if (tables[i].ttype)
// 	   free( tables[i].ttype);
// 	 if (tables[i].tdesc)
// 	   free( tables[i].tdesc);
//        }
	 
//        return OK;       
// }




// SQLQuery::run_update(char * query, timespec& ttime)

// ClientBaseAdt** tuple;
// BaseAdtInfo::BaseAttrType attrType;
// char * bFilename = NULL;
// int ret, err;
// int timeModIdx = 6;
// int count = 0;

// strcpy(queryStr, query);
// if(open_db() != OK) return DB_OPEN_ERROR;
	
	
//     for(;;) {

// 		count++;
		
// 		INFO("Updating profiles, query: " << query);

// 		//  check error codes here rather than assert
// 		if( (err = _cursor->beginTransaction() ) != clNOERROR) return err;
// 		if( (err = _cursor->declare(queryStr, true) ) != clNOERROR) return err;
		
// 		err = _cursor->open();
// 		if(err != clNOERROR){

// 			perror("An error occured while opening cursor");
// 			cout << "Error: " << err<<" "<<cem[err].message << endl;

// 			if(err == 19 && count < 5) {
// 				cout << "Trying open again after 2 sec" << endl;
// 				sleep(2);
				
// 				err = _cursor->close();
// 				if(err != 20) {
// 				      cout << "Close error: " << cem[err].message << endl;
// 				      _cursor->close();
// 				      close_db();
// 				      return ( CURSOR_ERROR );
// 				}
// 				else{
// 				  cout << "SQL: The transaction was ABORTED"<<endl;
// 				}
// 				continue;
// 			}
// 			else{
// 			    _cursor->close();
// 			    close_db();
// 			    return (FAIL);
// 			}
// 		}
		

// 		if(err == clEND_TUPLES) {
// 			cout << "No tuples.. " << endl;
// 			_cursor->close();
// 			close_db();
// 			return FAIL;
// 		}
				

// 		// "INDEX:" << timeModIdx << endl;
// 		if( (err = _cursor->fetch() ) != clNOERROR) { 
// 		  cout << "Unable to fetch tuples.. " << endl; 
// 		  CLEANUP; 
// 		  return FAIL; 
// 		}

// 		CL_CALL(_cursor->getTuple(tuple));
// 		ClientBaseAdt *obj=NULL;

// 		// get the ts_last field
// 		err = _cursor->getObj(obj,attrType,timeModIdx);
// 		if(err) { 
// 		    cout << "Unable to get the time (int) object.. " << endl; 
// 		    _cursor->close();
// 		    close_db();
// 		    return (CURSOR_ERROR); 
// 		}

// 		// storing the time as an int, and updating it's value
// 		ClientInt4Adt *aPtr = (ClientInt4Adt *)obj;
		
// 		INFO("old time: "<<(int)aPtr->getIntVal());
// 		aPtr->setIntVal(ttime.tv_sec);
// 		INFO("new time: "<<(int)aPtr->getIntVal());
		
// 		int modIdcs[1];
// 		modIdcs[0] = timeModIdx;
		
// 		cout <<endl<<endl;
// 		INFO("Attempting an update of field: "<<modIdcs[0]);
// 		err = _cursor->update(tuple, 1, modIdcs);

// 		if(err == 19 && count <= 50){
// 		  WARN("Update failed: " << cem[err].message << "Trying again in 5 seconds, count is: "<< count);
// 		  cout <<endl<<endl;	
// 		  sleep(2);
// 		  continue;
// 		}

// 		if(err) {
// 			WARN("Paradise error: " << cem[err].message );
// 			_cursor->close();
// 			close_db();
// 			return FAIL;
// 		}
		
// 		else {
// 			cout << "Update succeeded! Count: " << count << endl;
// 			break;
// 		}
// 	}

// 	if( _cursor->close() != OK) return CURSOR_ERROR;
// 	if( _cursor->commitTransaction() != OK ) return COMMIT_FAILED;
// 	if( close_db() != OK)  return DB_CLOSE_ERROR;
// 	return OK;
//}















