
#include <baddTranslator.h>

//----------------------------------------------------------------------------------------
//  constructor() - constructor for a badd query translator  
//
//  ARGS:    _profileId - the profile id
//           _clientId - the client id of client owning the profile
//           query - the query to be translated
//
//  RETURN:  none
//----------------------------------------------------------------------------------------
baddTranslator::baddTranslator(int _profileId, int _clientId, char * query)
{
    // initialize profile fields
    profileId = _profileId;
    clientId = _clientId;
    
    //---------------------------------------------------------------------------
    //  IMPORTANT: setup the scanner input buffer and load it with this query
    //  string.  Set current 'scanInput' pointer to the beg. of this input buffer
    //
    //  The following are global, defined in baddTranslator.h
    //  inbuf - the scanner buffer 
    //  scanInput - a pointer to the start of that buffer
    //  scanInputLimit - the size of the scanner buffer
    //---------------------------------------------------------------------------
    
    setQuery(query);

    int len = strlen( query );
    memcpy( inbuf, query, len );
    inbuf[len]='\0';
    scanInput = inbuf;
    inbuf[scanInputLimit-1]='\0';
}

baddTranslator::baddTranslator(int _profileId, int _clientId)
{
    // initialize profile fields
    profileId = _profileId;
    clientId = _clientId;
}

//----------------------------------------------------------------------------------------
//  translateQuery() - Translate the query and returns the translated query and number of 
//                     base table to the caller.
//  
//  ARGS:    status (by ref) - used to pass back an error code  { PARSE_FAILED, OK }
//
//  RETURN:  the translated query in a null terminated char array
//----------------------------------------------------------------------------------------
void
baddTranslator::setQuery(const char * query)
{
    int len = strlen( query );
    memcpy( inbuf, query, len );
    inbuf[len]='\0';
    scanInput = inbuf;
    inbuf[scanInputLimit-1]='\0';
    return;
}


//----------------------------------------------------------------------------------------
//  destructor()
//
//  ARGS:    none
//  RETURN:  none
//----------------------------------------------------------------------------------------
baddTranslator::~baddTranslator()
{ }


//----------------------------------------------------------------------------------------
//  translateQuery() - Translate the query and returns the translated query and number of 
//                     base table to the caller.
//  
//  ARGS:    status (by ref) - used to pass back an error code  { PARSE_FAILED, OK }
//
//  RETURN:  the translated query in a null terminated char array
//----------------------------------------------------------------------------------------
char * 
baddTranslator::translateQuery(int &status, int& numTemp)
{
    
    int      token, len, i;
    char     whereClause[1024]; 
    whereClause[0]=':';
    bool     success;
    
    status = OK;  // be optimistic
    
    // scan the buffer and strip off the where clause
    while ( (token = whereyylex()) ) {       
	switch (token) {
	case WHERE_CLAUSE:
	    //if(whereClause[0]!='\0'){
	        cerr<<"scanned the where clause: "<<whereyylval.t.strval;
		memcpy(whereClause, whereyylval.t.strval, 1024);
		// }
	    break;
	    
	case PRE_WHERE_CLAUSE:
	    /* nothing to do, continue */ 
	    cerr << "non where clause scanned"<<endl;
	    break;  
	    
	default:
	    cerr << "\tILLEGAL TOKEN!!"<<endl; flush(cerr);
	}
    }
    
    //----------------------------------------------------------------------
    // We now have the where clause (whereClause) and the original
    // query (inbuf).  If there is no where clause, whereClause[0] is NULL 
    // Now parse, typecheck and translate.  If parsing or typechecking
    // fail, set status and return;
    //----------------------------------------------------------------------
    
    // Set up Navin's opt++ parser stuff
    Acat_t catalog;
    Bparser_state_t parser_state;
    Aglob_vars()->cat = &catalog;
    Aglob_vars()->parser_state = &parser_state;
    
    
    // Set up the dummy catalog.  This is a hack at this point, just opens
    // a dummy file that the opt++ code can access as if it were the
    // Paradise catalogs.  Would be used in typechecking if typechecking
    // were done here, but it is not.
    ifstream catfile (".catalog");
    assert (!!catfile);
    catalog.read (catfile);
    
    //---------------
    //  Parse
    //---------------
    scanInput = inbuf;
    success = parseQuery(scanInput);     
    if(!success){
	status = PARSE_FAILED;
	return NULL;
    }
    
    
    // typecheck, no typechecking is done at this point, this should be added
    //success = typecheckQuery(inbuf);
    //if(!success){
    //  status = TYPECHECK_FAILED;
    //  return NULL;
    //}
    
    // get the number of base tables from the parse tree
    int numBase = Bparser_state()->query.arg(0).arg(1).nargs();
    
    char* baseNames[numBase];
    char* instIdents[numBase];
    
    const char* tmp=NULL;
    const char* tmp2=NULL;
    
    int numArgs;
    
    //------------------------------------------------------------
    // Get all base table names and idents from the parse tree
    //------------------------------------------------------------
    for( i =0; i<numBase; i++){  
	baseNames[i]=NULL;
	instIdents[i]=NULL;
	
	// there should be either 1 or 2 args.  1 if no instance ident, 
	// two if instance ident is used
	numArgs=Bparser_state()->query.arg(0).arg(1).arg(i).nargs();
	
	// get the insance identifier for this base table
	if(numArgs==2){
	    tmp2=Bparser_state()->query.arg(0).arg(1).arg(i).arg(1).text();
	    instIdents[i] = new char[256];
	    strcpy(instIdents[i], tmp2);
	}
	
	// get the base table name
	tmp=Bparser_state()->query.arg(0).arg(1).arg(i).arg(0).text();
	baseNames[i] = new char[256];
	strcpy(baseNames[i], tmp);
    }
    
    //----------------------------------------
    //  Generate the translated query string
    //----------------------------------------
    char* translated_query = constructQuery(numBase, baseNames, 
					    instIdents, whereClause);
    
    //INFO( "translated query: " <<translated_query );
    
    //------------------------------
    //  Cleanup
    //------------------------------    
    for( i =0; i<numBase; i++){  
	if( baseNames[i] )
	    delete[] baseNames[i];
	if( instIdents[i] )
	    delete[] instIdents[i];
    }
    
    numTemp=numBase;
    return translated_query;     
}



//----------------------------------------------------------------------------------------
//  parseQuery() -  Parse the query with Navin's opt++ parser, return true if parsed, false 
//                  if parse fails. If parsed ok, the parse tree is in the global variable 
//                  object as " Bparser_state()->query "
//
//  ARGS:    The query to parse
//
//  RETURN:  The error code { true, false }
//----------------------------------------------------------------------------------------
bool 
baddTranslator::parseQuery(char* inQuery)
{  
    Bparser_state()->input_string = inQuery;  
    
    extern void Brestart_lexical_analyzer (void);
    Brestart_lexical_analyzer ();
    extern int yyparse (void);
    
    if (yyparse ()){
	return false;
    }
    return true;
}



//----------------------------------------------------------------------------------------
//  typecheckQuery() -  This method should take the original query and typecheck it.
//                      Nothing is done for now, but typechecking could be added 
//
//  ARGS:    the query to typecheck
//
//  RETURN:  error code { true , false }
//----------------------------------------------------------------------------------------
bool 
baddTranslator::typecheckQuery(char* inQuery)
{
    return true;
}


//----------------------------------------------------------------------------------------
//  constructQuery() -  Takes the original whereClause and information obtained from the
//                      parse tree and constructs the badd translated sql string. 
//                      This translated string is returned to caller
//
//  ARGS:    numBase - the number of bases tables in the original query
//           names - the names of those base tables
//           idents - the identifiers used for those base tables
//           originalWhere - the where clause from the original query
//
//  RETURN:  the translated query, or NULL if translation fails
//----------------------------------------------------------------------------------------
char * 
baddTranslator::constructQuery(int numBase, char** names, char** idents, 
			       char* originalWhere)
{
    
    ostrstream insertClause, selectClause, fromClause;
    ostrstream whereClause, finalQuery, boolStmt;
    char* tmp;
    int i;    
    
    insertClause << "INSERT INTO T" << clientId<<"_"
		 <<profileId<<"_R "<<ends;
    
    //------------------------------------------------------------
    // Generate select clause: select the timestamps and ids from 
    // each base table if there is an ident used for this table 
    // name, use ident in query.  Else use the table name
    //------------------------------------------------------------
    selectClause << "SELECT ";
    for(i=0; i<numBase; i++){
	if(idents[i]!=NULL)
	    tmp=idents[i];
	else
	    tmp=names[i];
	selectClause << tmp<<".TS, " << tmp<<".ID, ";
    }
    
    ////////////////////////////////////////////////////////////
    // 
    // Create the boolean 'old' expression based on profile.TS
    // and the TS of insertion for each joining tuple.  There 
    // are 2 cases to deal with: 
    //
    //   1) a query with a single base table need only use 
    //      the lessThan operator
    //   2) queries with multiple base tables need to generate 
    //      a statement with nested booleanAnd operations
    //
    ////////////////////////////////////////////////////////////
    if(numBase==1){
	if(idents[0]!=NULL) 
	    tmp=idents[0];
	else
	    tmp=names[0];
	boolStmt << "LESSTHAN("<<tmp<<".TS, pr.TS) "<<ends;
    }
    
    if(numBase>=2){
	int numLeft=numBase-1;
	
	while(numLeft>0){
	    // add the next nested level to the boolean condition
	    if(idents[numLeft]==NULL && names[numLeft]==NULL)
		numLeft++;
	    if(idents[numLeft]!=NULL) 
		tmp=idents[numLeft];
	    else
		tmp=names[numLeft];
	    boolStmt << "BOOLEANAND( LESSTHAN("<<tmp<<".TS, pr.TS), ";
	    numLeft--;
	}      
	
	if(idents[0]!=NULL) tmp=idents[0];
	else                tmp=names[0];
	boolStmt << "LESSTHAN("<<tmp<<".TS, pr.TS) ";
	for(int t=0; t<numBase-1; t++)
	    boolStmt << ")";
	boolStmt << " "<<ends;
    }
    
    
    // add the bool stmt to the select clause
    selectClause << boolStmt.str()<<ends;
    
    //------------------------------------------------------------
    // Generate the from clause. If idents are used, use the 
    // idents in the from clause, else use base table names
    //------------------------------------------------------------
    fromClause << "FROM ";  
    for(i=0; i<numBase; i++){
	fromClause << names[i]; 
	if( idents[i]!=NULL )
	    fromClause << " "<<idents[i];
	fromClause << ", ";
    }
    fromClause << "PROFILES PR "<<ends;

    //------------------------------------------------------------
    // Generate the where clause, use original
    // where clause and add profile conditions
    //------------------------------------------------------------
    if(originalWhere[0]!= '\0'){
	whereClause << originalWhere<<" AND PR.ID="<<profileId
		    <<" AND PR.CID="<<clientId<<" "<<ends;
    }
    else{
	whereClause << "WHERE PR.ID="<< profileId
		    <<" AND PR.CID="<< clientId<<" "<<ends;
    }
    
    // Add each clause to the final query string
    // This is the first temp query for this profile
    finalQuery << insertClause.str() << selectClause.str() 
	       << fromClause.str()<< whereClause.str()<<" ; ";
    
    
    
    /////////////////////////////////////////////////////////////
    // Generate the second tmp queries to fill each temp table
    // associated with each base table, queries are of the form:
    // for client 101, profile 2
    //
    //     INSERT into t101_2_1 SELECT tmp1.B1ID, min(tmp1.old)  
    //     FROM t101_2_R tmp1 GROUP BY tmp1.B1ID;  
    //
    /////////////////////////////////////////////////////////////
    ostrstream queries[numBase];
    
    for(i=0; i<numBase; i++){
	queries[i]<< "INSERT INTO T"<<clientId
		  << "_"<<profileId<<"_"<<i+1
		  << " SELECT TMP1.B"<<i+1<<"ID, MIN(TMP1.ISOLD) "
		  << "FROM T"<<clientId <<"_"
		  << profileId <<"_R TMP1 " 
		  << "GROUP BY TMP1.B"<<i+1<<"ID"<<ends;
	
	finalQuery << queries[i].str()<<" ; ";
	
	// Reset the streams here, they will be used below to 
	// create the final queries    
	queries[i].freeze(0);
	queries[i].seekp(0);
    }
    
    
    finalQuery << " $ ";   // delimiter for distinguishing setup queries 
    // from dissemination queries
    
    /////////////////////////////////////////////////////////////
    // Generate the final queries. These queries will produce the 
    // base table tuples to disseminate to the client.  Queries
    // are of the following form for an example client 101, profile 2
    // and basetable named 'landings'
    //
    //      SELECT * FROM landings, t101_2_1 t WHERE landings.ID=t.B1ID 
    //      AND t.minOfOld=false
    //
    /////////////////////////////////////////////////////////////
    ostrstream final_queries[numBase];       
    for(i=0; i<numBase; i++){
	final_queries[i]<< "SELECT * FROM "<<names[i]<<", T"<<clientId
			<<"_"<< profileId<<"_"<<i+1<<" TMP2 WHERE "<<names[i]
			<< ".ID=TMP2.B"<<i+1<<"ID AND TMP2.MINOFOLD=FALSE"<<ends;
	finalQuery << final_queries[i].str()<<" ;";
	final_queries[i].freeze(0);
    }
    
    finalQuery << ends;
    insertClause.freeze(0);
    selectClause.freeze(0);
    fromClause.freeze(0);
    whereClause.freeze(0);
    boolStmt.freeze(0);
    
    return finalQuery.str();
}










